# -*- coding: utf-8 -*-

# Copyright © 2012-2017 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function, unicode_literals

import sys
import copy

from Onboard.Version        import require_gi_versions
require_gi_versions()
from gi.repository          import Gdk

from Onboard.utils          import EventSource, unicode_str
from Onboard.definitions    import UINPUT_DEVICE_NAME

import Onboard.osk as osk

import logging
_logger = logging.getLogger(__name__)


class XIEventType:
    """ enum of XInput events """
    DeviceChanged     = 1
    KeyPress          = 2
    KeyRelease        = 3
    ButtonPress       = 4
    ButtonRelease     = 5
    Motion            = 6
    Enter             = 7
    Leave             = 8
    FocusIn           = 9
    FocusOut          = 10
    HierarchyChanged  = 11
    PropertyEvent     = 12
    RawKeyPress       = 13
    RawKeyRelease     = 14
    RawButtonPress    = 15
    RawButtonRelease  = 16
    RawMotion         = 17
    TouchBegin        = 18
    TouchUpdate       = 19
    TouchEnd          = 20
    TouchOwnership    = 21
    RawTouchBegin     = 22
    RawTouchUpdate    = 23
    RawTouchEnd       = 24

    # extra non-XI events
    DeviceAdded       = 1100
    DeviceRemoved     = 1101
    SlaveAttached     = 1102
    SlaveDetached     = 1103

    HierarchyEvents = (DeviceAdded, DeviceRemoved, SlaveAttached, SlaveDetached)


class XIEventMask:
    """ enum of XInput event masks """
    DeviceChangedMask     = 1 << XIEventType.DeviceChanged
    KeyPressMask          = 1 << XIEventType.KeyPress
    KeyReleaseMask        = 1 << XIEventType.KeyRelease
    ButtonPressMask       = 1 << XIEventType.ButtonPress
    ButtonReleaseMask     = 1 << XIEventType.ButtonRelease
    MotionMask            = 1 << XIEventType.Motion
    EnterMask             = 1 << XIEventType.Enter
    LeaveMask             = 1 << XIEventType.Leave
    FocusInMask           = 1 << XIEventType.FocusIn
    FocusOutMask          = 1 << XIEventType.FocusOut
    HierarchyChangedMask  = 1 << XIEventType.HierarchyChanged
    PropertyEventMask     = 1 << XIEventType.PropertyEvent
    RawKeyPressMask       = 1 << XIEventType.RawKeyPress
    RawKeyReleaseMask     = 1 << XIEventType.RawKeyRelease
    RawButtonPressMask    = 1 << XIEventType.RawButtonPress
    RawButtonReleaseMask  = 1 << XIEventType.RawButtonRelease
    RawMotionMask         = 1 << XIEventType.RawMotion
    TouchBeginMask        = 1 << XIEventType.TouchBegin
    TouchUpdateMask       = 1 << XIEventType.TouchUpdate
    TouchEndMask          = 1 << XIEventType.TouchEnd
    TouchOwnershipMask    = 1 << XIEventType.TouchOwnership
    RawTouchBeginMask     = 1 << XIEventType.RawTouchBegin
    RawTouchUpdateMask    = 1 << XIEventType.RawTouchUpdate
    RawTouchEndMask       = 1 << XIEventType.RawTouchEnd

    TouchMask             = TouchBeginMask | \
                            TouchUpdateMask | \
                            TouchEndMask

    RawTouchMask          = RawTouchBeginMask | \
                            RawTouchUpdateMask | \
                            RawTouchEndMask

class XIDeviceType:
    """ enum of XInput device types """
    MasterPointer  = 1
    MasterKeyboard = 2
    SlavePointer   = 3
    SlaveKeyboard  = 4
    FloatingSlave  = 5


class XITouchMode:
    DirectTouch    = 1
    DependentTouch = 2


class XIDeviceManager(EventSource):
    """
    XInput device manager singleton.
    """

    blacklist = ("Virtual core XTEST keyboard",
                 UINPUT_DEVICE_NAME,
                 "Power Button")
    last_device_blacklist = ("Virtual core XTEST pointer")

    def __new__(cls, *args, **kwargs):
        """
        Singleton magic.
        """
        if not hasattr(cls, "self"):
            cls.self = object.__new__(cls, *args, **kwargs)
            cls.self.construct()
        return cls.self

    def __init__(self):
        """
        Called multiple times, do not use. In particular don't
        call base class constructors here.
        """
        pass

    def construct(self):
        """
        Singleton constructor, runs only once.
        """
        EventSource.__init__(self, ["device-event", "device-grab",
                                    "devices-updated"])

        self._devices = {}
        self._osk_devices = None
        try:
            self._osk_devices = osk.Devices(event_handler = \
                                            self._on_device_event)
        except Exception as ex:
            _logger.warning("Failed to create osk.Devices: " + \
                            unicode_str(ex))

        self._last_motion_device_id = None
        self._last_click_device_id = None
        self._last_device_blacklist_ids = []

        self._grabbed_devices_ids = set()

        if self.is_valid():
            self.update_devices()

    def is_valid(self):
        return not self._osk_devices is None

    def lookup_device_id(self, device_id):
        return self._devices.get(device_id)

    def lookup_config_string(self, device_config_string):
        for device in self.get_pointer_devices():
            if device.get_config_string() == device_config_string:
                return device

    def get_client_pointer(self):
        """ Return client pointer device """
        device_id = self._osk_devices.get_client_pointer()
        return self.lookup_device_id(device_id)

    def get_client_keyboard(self):
        """ Return client keyboard device """
        client_pointer = self.get_client_pointer()
        device_id = client_pointer.attachment
        return self.lookup_device_id(device_id)

    def get_devices(self):
        return self._devices.values()

    def get_pointer_devices(self):
        return [device for device in self._devices.values() \
                if device.is_pointer()]

    def get_client_pointer_slaves(self):
        """
        All slaves of the client pointer, with and without device grabs.
        """
        client_pointer = self.get_client_pointer()
        return [device for device in self.get_pointer_devices() \
                if not device.is_master() and \
                   device.attachment == client_pointer.id]

    def get_client_pointer_attached_slaves(self):
        """
        Slaves that are currently attached to the client pointer.
        """
        devices = self.get_client_pointer_slaves()
        devices = [d for d in devices if not d.is_floating() and \
                                         not self.is_grabbed(d)]
        return devices

    def get_keyboard_devices(self):
        return [device for device in self._devices.values() \
                if device.is_keyboard()]

    def get_client_keyboard_slaves(self):
        """
        All slaves of the keyboard paired with the client pointer,
        with and without device grabs.
        """
        client_keyboard = self.get_client_keyboard()
        return [device for device in self.get_keyboard_devices() \
                if not device.is_master() and \
                   device.attachment == client_keyboard.id]

    def get_client_keyboard_attached_slaves(self):
        """
        Slaves that are currently attached to the keyboard paired
        with the client pointer.
        """
        devices = self.get_client_keyboard_slaves()
        devices = [d for d in devices if not d.is_floating() and \
                                         not self.is_grabbed(d)]
        return devices

    def get_master_pointer_devices(self):
        return [device for device in self.get_pointer_devices() \
                if device.is_master()]

    def update_devices(self):
        devices = {}
        self._last_device_blacklist_ids = []

        for info in self._osk_devices.list():
            device = XIDevice()
            device._device_manager = self
            (
                device.name,
                device.id,
                device.use,
                device.attachment,
                device.enabled,
                device.vendor,
                device.product,
                device.touch_mode,
            ) = info
            device.source = XIDevice.classify_source(device.name, device.use,
                                                     device.touch_mode)

            if sys.version_info.major == 2:
                device.name = unicode_str(device.name)

            if not device.name in self.blacklist:
                devices[device.id] = device

            if device.name in self.last_device_blacklist:
                self._last_device_blacklist_ids.append(device.id)

        self._devices = devices

        self._last_click_device_id = None
        self._last_motion_device_id = None

        # notify listeners about the previous devices becoming invalid
        self.emit("devices-updated")

    def select_events(self, window, device, mask):
        if window is None:  # use root window?
            xid = 0
        else:
            xid = window.get_xid()
            if not xid:
                return False

        self._osk_devices.select_events(xid, device.id, mask)
        return True

    def unselect_events(self, window, device):
        if window is None:  # use root window?
            xid = 0
        else:
            xid = window.get_xid()
            if not xid:
                return False

        self._osk_devices.unselect_events(xid, device.id)
        return True

    def attach_device(self, device, master_id):
        self.attach_device_id(device.id, master_id)

    def detach_device(self, device):
        self.detach_device_id(device.id)

    def attach_device_id(self, device_id, master_id):
        self._osk_devices.attach(device_id, master_id)

    def detach_device_id(self, device_id):
        self._osk_devices.detach(device_id)

    def grab_device(self, device):
        self.grab_device_id(device.id) # raises osk.error
        self.emit("device-grab", device, True)

    def ungrab_device(self, device):
        self.ungrab_device_id(device.id) # raises osk.error
        self.emit("device-grab", device, False)

    def is_grabbed(self, device):
        return device.id in self._grabbed_devices_ids

    def grab_device_id(self, device_id):
        self._osk_devices.grab_device(device_id, 0)

        assert(not device_id in self._grabbed_devices_ids)
        self._grabbed_devices_ids.add(device_id)

    def ungrab_device_id(self, device_id):
        self._grabbed_devices_ids.discard(device_id)
        self._osk_devices.ungrab_device(device_id)

    def get_last_click_device(self):
        id = self._last_click_device_id
        if id is None:
            return None
        return self.lookup_device_id(id)

    def get_last_motion_device(self):
        id = self._last_motion_device_id
        if id is None:
            return None
        return self.lookup_device_id(id)

    def _on_device_event(self, event):
        """
        Handler for XI2 events.
        """
        event_type = event.xi_type
        device_id = event.device_id
        source_id = event.source_id

        # update our device objects on changes to the device hierarchy
        if event_type in XIEventType.HierarchyEvents or \
           event_type == XIEventType.DeviceChanged:
            self.update_devices()

        # simulate gtk source device
        if source_id:
            source_device = self.lookup_device_id(source_id)
            if not source_device:
                return
        else:
            source_device = None
        event.set_source_device(source_device)

        ## debug, simulate touch-screen
        if 0 and \
           device_id == 11:
            if not self._disguise_as_touch_event(event, False):
                return

        # remember recently used device ids for CSFloatingSlave
        if not source_id in self._last_device_blacklist_ids:
            if event_type == XIEventType.Motion:
                self._last_motion_device_id = source_id
            elif event_type == XIEventType.ButtonPress or \
                 event_type == XIEventType.TouchBegin:
                self._last_click_device_id = source_id

        # forward the event to all listeners
        for callback in self._callbacks["device-event"]:
            # Copy event to isolate callbacks from each other (LP: 1421840)
            ev = copy.copy(event)
            callback(ev)

    def _disguise_as_touch_event(self, event, wacom_mode=False):
        """
        Disguise a pointer event as touch event.
        For debugging purposes only.

        Set wacom_mode=True to simulate pointer events coming
        from a "touch screen" device that only sends pointer events.
        This is the default case for wacom touch screens due to
        gestures being enabled, see LP #1297692.
        """
        device = self.lookup_device_id(event.source_id)
        device.name         = "Touch-Screen"
        device.source       = Gdk.InputSource.TOUCHSCREEN
        device.touch_mode   = XITouchMode.DirectTouch

        if event.xi_type == XIEventType.Motion and \
           not event.state & (Gdk.ModifierType.BUTTON1_MASK | \
                                Gdk.ModifierType.BUTTON2_MASK | \
                                Gdk.ModifierType.BUTTON3_MASK):
            return False  # discard event

        if not wacom_mode:
            if event.xi_type == XIEventType.ButtonPress:
                event.xi_type = XIEventType.TouchBegin
                event.type = Gdk.EventType.TOUCH_BEGIN
            if event.xi_type == XIEventType.ButtonRelease:
                event.xi_type = XIEventType.TouchEnd
                event.type = Gdk.EventType.TOUCH_END
            if event.xi_type == XIEventType.Motion:
                event.xi_type = XIEventType.TouchUpdate
                event.type = Gdk.EventType.TOUCH_UPDATE

            event.sequence = 10  # single touch only

        event.button = 1

        return True  # allow event


class XIDevice(object):
    """
    XInput device wrapper.
    """
    name         = None
    id           = None
    use          = None
    attachment   = None  # master for slaves, paired master for masters
    enabled      = None
    vendor       = None
    product      = None
    source       = None
    touch_mode   = None

    _device_manager = None

    def __repr__(self):
        return "{}(id={}, attachment={}, name={}, source={} )" \
                .format(type(self).__name__,
                        repr(self.id),
                        repr(self.attachment),
                        repr(self.name),
                        repr(self.source),
                       )

    def __str__(self):
        return "{}(id={} attachment={} use={} touch_mode={} source={} name={} " \
               "vendor=0x{:04x} product=0x{:04x} enabled={})" \
                .format(type(self).__name__,
                        self.id,
                        self.attachment,
                        self.use,
                        self.touch_mode,
                        self.get_source().value_name,
                        self.name,
                        self.vendor,
                        self.product,
                        self.enabled,
                       )

    def get_source(self):
        """
        Return Gdk.InputSource for compatibility with Gtk event handling.
        """
        return self.source

    @staticmethod
    def classify_source(name, use, touch_mode):
        """
        Determine the source type (Gdk.InputSource) of the device.
        Logic taken from GDK, gdk/x11/gdkdevicemanager-xi2.c
        """
        if use == XIDeviceType.MasterKeyboard or \
           use == XIDeviceType.SlaveKeyboard:
            input_source = Gdk.InputSource.KEYBOARD
        elif use == XIDeviceType.SlavePointer and \
            touch_mode:
            if touch_mode == XITouchMode.DirectTouch:
                input_source = Gdk.InputSource.TOUCHSCREEN
            else:
                input_source = Gdk.InputSource.TOUCHPAD
        else:
            name = unicode_str(name.lower())
            if "eraser" in name:
                input_source = Gdk.InputSource.ERASER
            elif "cursor" in name:
                input_source = Gdk.InputSource.CURSOR
            elif "wacom" in name or \
                 "pen" in name:   # uh oh, false positives?
                input_source = Gdk.InputSource.PEN
            else:
                input_source = Gdk.InputSource.MOUSE
        return input_source

    def is_touch_screen(self):
        """
        Touch screen device?
        """
        return self.source == Gdk.InputSource.TOUCHSCREEN

    # methods inherited from Gerd's scanner device.
    def is_master(self):
        """
        Is this a master device?
        """
        return self.use == XIDeviceType.MasterPointer or \
               self.use == XIDeviceType.MasterKeyboard

    def is_pointer(self):
        """
        Is this device a pointer?
        """
        return self.use == XIDeviceType.MasterPointer or \
               self.use == XIDeviceType.SlavePointer

    def is_keyboard(self):
        """
        Is this device a keyboard?
        """
        return self.use == XIDeviceType.MasterKeyboard or \
               self.use == XIDeviceType.SlaveKeyboard

    def is_floating(self):
        """
        Is this device detached?
        """
        return self.use == XIDeviceType.FloatingSlave

    def get_config_string(self):
        """
        Get a configuration string for the device.
        Format: VID:PID:USE

        """
        return "{:04X}:{:04X}:{!s}".format(self.vendor,
                                           self.product,
                                           self.use)

class XIDeviceEventLogger:
    """
    Facilities for logging device events.
    Has little overhead when logging is disabled.
    """

    def __init__(self):
        if not _logger.isEnabledFor(logging.DEBUG):
            self.log_event = self._log_event_stub

    def _log_device_event(self, event):
        if not event.xi_type in [ XIEventType.TouchUpdate,
                                  XIEventType.Motion]:
            self.log_event("Device event: dev_id={} src_id={} xi_type={} "
                          "xid_event={}({}) x={} y={} x_root={} y_root={} "
                          "button={} state={} sequence={}"
                          "".format(event.device_id,
                                    event.source_id,
                                    event.xi_type,
                                    event.xid_event,
                                    self.get_xid(),
                                    event.x, event.y,
                                    event.x_root, event.y_root,
                                    event.button, event.state,
                                    event.sequence,
                                   )
                         )

            device = event.get_source_device()
            self.log_event("Source device: " + str(device))

    @staticmethod
    def log_event(msg, *args):
        _logger.event(msg.format(*args))

    @staticmethod
    def _log_event_stub(msg, *args):
        pass





# Copyright © 2009 Chris Jones <tortoise@tortuga>
# Copyright © 2012-2013 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

PYTHON_EXECUTABLE = "python3"

from gi.repository import GLib

def run():
    GLib.spawn_async(argv=[PYTHON_EXECUTABLE,
                           "-cfrom Onboard.settings import Settings\ns = Settings(False)"],
                     flags=GLib.SpawnFlags.SEARCH_PATH)




# -*- coding: utf-8 -*-

# Copyright © 2012-2017 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" GTK specific keyboard class """

from __future__ import division, print_function, unicode_literals

import os
import time
import logging
_logger = logging.getLogger(__name__)

from Onboard.Version   import require_gi_versions
require_gi_versions()
try:
    from gi.repository import Atspi
except ImportError as e:
    _logger.warning("Atspi typelib missing, auto-show unavailable")

from Onboard.utils     import Rect, EventSource, Process, unicode_str
from Onboard.Timer     import Timer

# Config Singleton
from Onboard.Config import Config
config = Config()


class CachedAccessible:
    def __init__(self, accessible):
        self._accessible = accessible
        self._state = {}       # cache of various accessible properties

    # Use "==" for object identity tests instead of "is".
    def __eq__(self, other):
        return other is not None and self._accessible is other._accessible

    def __ne__(self, other):
        return other is None or self._accessible is not other._accessible

    def get_state(self):
        """ All cached state of the accessible """
        return self._state

    def get_all_state(self):
        """
        Return _state filled with all kinds of properties, for easy printint
        as debug output in TextContext.
        """
        self.get_role()
        self.get_role_name()
        self.get_name()
        self.get_state_set()
        self.get_id()
        self.get_attributes()
        self.get_interfaces()
        self.get_description()
        self.get_pid()
        self.get_process_name()
        self.get_toolkit_name()
        self.get_toolkit_version()
        self.get_editable_text_iface()
        self.get_app_name()
        self.get_app_description()
        self.get_app_toolkit_name()
        self.get_extents()
        self.get_frame()
        self.get_frame_extents()
        self.is_urlbar()
        self.is_byobu()
        return self._state

    # ### Cached, exception-safe accessor functions ###

    def get_role(self):
        return self._get_value("role",
                               self._accessible.get_role)

    def get_role_name(self):
        return self._get_value("role-name",
                               self._accessible.get_role_name)

    def get_name(self):
        return self._get_value("name",
                               self._accessible.get_name)

    def invalidate_state_set(self):
        self.invalidate("state-set")

    def get_state_set(self):
        return self._get_value("state-set",
                               self._accessible.get_state_set)

    def get_id(self):
        return self._get_value("id",
                               self._accessible.get_id)

    def get_attributes(self):
        return self._get_value("attributes",
                               self._accessible.get_attributes, {})

    def get_interfaces(self):
        return self._get_value("interfaces",
                               self._accessible.get_interfaces, [])

    def get_description(self):
        return self._get_value("description",
                               self._accessible.get_description)

    def get_pid(self):
        return self._get_value("pid",
                               self._accessible.get_process_id)

    def get_process_name(self):
        pid = self.get_pid()
        if pid != -1:
            return self._get_value_noex(
                "process-name",
                lambda : Process.get_process_name(pid))
        return None

    def get_toolkit_name(self):
        return self._get_value("toolkit-name",
                               self._accessible.get_toolkit_name)

    def get_toolkit_version(self):
        return self._get_value("toolkit-version",
                               self._accessible.get_toolkit_version)

    def get_editable_text_iface(self):
        try:
            func = self._accessible.get_editable_text_iface
        except AttributeError:  # function does not exist in Precise
            return None

        return self._get_value("editable-text-iface", func)

    def get_app_name(self):
        def func():
            app = self.get_application()
            return app.get_name() if app else ""

        return self._get_value("app-name", func, "")

    def get_app_description(self):
        def func():
            app = self.get_application()
            return app.get_description() if app else ""

        return self._get_value("app-description", func, "")

    def get_app_toolkit_name(self):
        """
        Toolkit name reported by the *application* accessible.

        AT-SPI canonically reports the toolkit on the application root
        rather than on each widget; for Qt apps (Kate, KWrite, Konsole,
        Dolphin, etc.) the widget's own get_toolkit_name() returns
        None while the application's returns "Qt". GTK apps mirror
        their toolkit on widget attributes; Qt apps don't, hence this
        separate accessor.
        """
        def func():
            app = self.get_application()
            try:
                return app.get_toolkit_name() if app else ""
            except Exception:
                return ""
        return self._get_value("app-toolkit-name", func, "")

    def get_application(self):
        return self._get_value("application",
                               self._accessible.get_application)

    def invalidate_extents(self):
        self.invalidate("extents")

    def get_extents(self):
        """
        Screen rect after scaling.
        """
        scale = config.window_scaling_factor
        if scale != 1.0:
            # Only Gtk-3 widgets return scaled coordinates, all others,
            # including Gtk-2 apps like firefox, clawsmail and Qt-apps,
            # apparently don't.
            if self.is_toolkit_gtk3():
                scale = 1.0
            else:
                scale = 1.0 / config.window_scaling_factor

        def func():
            ext = self._accessible.get_extents(Atspi.CoordType.SCREEN)
            return Rect(ext.x * scale, ext.y * scale,
                        ext.width * scale, ext.height * scale)

        return self._get_value("extents", func, Rect())

    def get_frame(self):
        def func():
            frame = self._get_accessible_frame(self._accessible)
            if frame:
                return CachedAccessible(frame)
            return None

        return self._get_value_noex("frame", func)

    def get_frame_extents(self):
        def func():
            frame = self.get_frame()
            if frame:
                return frame.get_extents()
            return Rect()

        return self._get_value_noex("frame_extents", func)

    @staticmethod
    def _get_accessible_frame(accessible):
        """ Accessible of the top level window to which accessible belongs. """
        frame = None
        _logger.atspi("_get_accessible_frame(): searching for top level:")
        try:
            parent = accessible
            while True:
                parent = parent.get_parent()
                if not parent:
                    break
                role = parent.get_role()
                _logger.atspi("parent: {}".format(role))
                if role == Atspi.Role.FRAME or \
                   role == Atspi.Role.DIALOG or \
                   role == Atspi.Role.WINDOW or \
                   role == Atspi.Role.NOTIFICATION:
                    frame = parent
                    break
        # private exception gi._glib.GError when
        # right clicking onboards unity2d launcher (Precise)
        except Exception as ex:
            _logger.atspi("Invalid accessible,"
                          " failed to get top level accessible: " +
                          unicode_str(ex))
        return frame

    def is_urlbar(self):
        """ Is this a (most likely firefox') URL bar? """
        def func():
            attributes = self.get_attributes()
            return bool(attributes and "urlbar" in attributes.get("class", ""))

        return self._get_value_noex("is_urlbar", func)

    def is_byobu(self):
        """ Is this possibly byobu running in a terminal? """
        def func():
            description = self.get_description()
            return bool(description and "byobu" in description.lower())

        return self._get_value_noex("is_byobu", func)

    def _get_value(self, name, func, default=None):
        """ Return cached return value of func(). """
        value = self._state.get(name)
        if value is None:
            try:
                value = func()
            except Exception as ex:  # private exception gi._glib.GError
                _logger.info("CachedAccessible._get_value({}): "
                             "invalid accessible, failed to read state: "
                             .format(name) + unicode_str(ex))
                value = default

            self._state[name] = value

        return value

    def _get_value_noex(self, name, func):
        """ Return cached return value of func(). """
        value = self._state.get(name)
        if value is None:
            value = func()
            self._state[name] = value
        return value

    def invalidate(self, name):
        """
        Force re-reading property from the accessible.
        May cause a D-Bus round-trip on the next read-attempt.
        """
        try:
            del self._state[name]
        except KeyError:
            pass

    # ### uncached, but still exception safe functions ###

    def get_selection(self, selection_num=0):
        selection = None
        try:
            sel = self._accessible.get_selection(selection_num)
            # Gtk-2 applications return 0,0 when there is no selection.
            # Gtk-3 applications return caret positions in that case.
            # LibreOffice Writer in Vivid initially returns -1,-1 when there
            # is no selection, later the caret position.
            start = sel.start_offset
            end = sel.end_offset
            if start > 0 and \
               end > 0 and \
               start <= end:
                selection = (sel.start_offset, sel.end_offset)
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.get_selection(): " +
                         unicode_str(ex))
        return selection

    def set_caret_offset(self, offset):
        try:
            self._accessible.set_caret_offset(offset)
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.set_caret_offset(): " +
                         unicode_str(ex))

    def insert_text(self, position, text):
        # AT-SPI's atspi_editable_text_insert_text() takes the number
        # of *characters* to insert, not bytes. Passing -1 works for
        # the GTK bridge (it computes strlen on its side) but Qt's
        # accessibility bridge truncates to length=0 on -1 and the
        # insert silently no-ops. Pass the actual character count.
        try:
            ok = self._accessible.insert_text(position, text, len(text))
            _logger.atspi(
                "CachedAccessible.insert_text(pos={}, text={!r}, "
                "len={}) -> {}".format(position, text, len(text), ok))
            return ok
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.warning("CachedAccessible.insert_text() failed: " +
                            unicode_str(ex))
        return False

    def delete_text(self, start_pos, end_pos):
        try:
            return self._accessible.delete_text(start_pos, end_pos)
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.delete_text(): " +
                         unicode_str(ex))
        return False

    # ### uncached, raising exceptions ###

    def get_caret_offset(self):
        try:
            offset = self._accessible.get_caret_offset()
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.get_caret_offset(): " +
                         unicode_str(ex))
            raise ex
        return offset

    def get_character_count(self):
        try:
            count = self._accessible.get_character_count()
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.get_character_count(): " +
                         unicode_str(ex))
            raise ex
        return count

    @staticmethod
    def _boundary_to_granularity(boundary_type):
        """
        Map deprecated Atspi.TextBoundaryType values to
        Atspi.TextGranularity equivalents used by get_string_at_offset().
        The _START/_END distinction has no granularity counterpart;
        both collapse to the same granularity bucket.
        """
        BT = Atspi.TextBoundaryType
        TG = Atspi.TextGranularity
        mapping = {
            BT.CHAR:           TG.CHAR,
            BT.WORD_START:     TG.WORD,
            BT.WORD_END:       TG.WORD,
            BT.SENTENCE_START: TG.SENTENCE,
            BT.SENTENCE_END:   TG.SENTENCE,
            BT.LINE_START:     TG.LINE,
            BT.LINE_END:       TG.LINE,
        }
        return mapping.get(boundary_type, TG.LINE)

    def _string_at(self, offset, boundary_type):
        """
        Call the modern Atspi.Text.get_string_at_offset(), or return
        None when the underlying accessible only exposes the deprecated
        get_text_at_offset() / get_text_before_offset() API. Callers
        check for None to decide whether to fall back to the legacy
        call appropriate for their context.
        """
        if not hasattr(self._accessible, "get_string_at_offset"):
            return None
        granularity = self._boundary_to_granularity(boundary_type)
        return self._accessible.get_string_at_offset(offset, granularity)

    def get_text_at_offset(self, offset, boundary_type):
        # Atspi.Text.get_text_at_offset() is deprecated and refused by
        # current at-spi2-core; prefer get_string_at_offset() with a
        # fallback to the legacy call for very old AT-SPI.
        try:
            text = self._string_at(offset, boundary_type)
            if text is None:
                text = self._accessible.get_text_at_offset(offset,
                                                           boundary_type)
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.get_text_at_offset(): " +
                         unicode_str(ex))
            raise ex
        return text

    def get_text_before_offset(self, offset, boundary_type):
        # GetStringAtOffset has no "before" variant; emulate by reading
        # the segment at the offset, then the segment just before it.
        try:
            cur = self._string_at(offset, boundary_type)
            if cur is None:
                text = self._accessible.get_text_before_offset(offset,
                                                               boundary_type)
            elif cur.start_offset > 0:
                text = self._string_at(cur.start_offset - 1, boundary_type)
            else:
                text = cur
                text.content = ""
                text.end_offset = 0
        except Exception as ex:  # Private exception gi._glib.GErro
            _logger.info("CachedAccessible.get_text_before_offset(): " +
                         unicode_str(ex))
            raise ex
        return text

    def get_text(self, begin, end):
        """ Text of the given accessible, no caching """
        try:
            text = Atspi.Text.get_text(self._accessible, begin, end)
        # private exception gi._glib.GError: timeout from dbind
        # with web search in firefox.
        except Exception as ex:
            _logger.atspi("CachedAccessible.get_text(): " +
                          unicode_str(ex))
            raise ex
        return text

    # ### Higher level functions ###

    def is_focused(self, invalidate=False):
        if invalidate:  # re-read properties?
            self.invalidate_state_set()

        state_set = self.get_state_set()
        if state_set is not None:
            return state_set.contains(Atspi.StateType.FOCUSED)
        return False

    def is_editable(self):
        """ Is this an accessible onboard should be shown for? """
        state_set = self.get_state_set()
        if state_set and state_set.contains(Atspi.StateType.EDITABLE):
            return True
        role      = self.get_role()
        return role in [
                    Atspi.Role.TERMINAL,
                    ]

    def is_not_focus_stealing(self):
        """
        Is this accessible unlikely to steal the focus from
        a previously focused editable accessible?
        """
        role      = self.get_role()
        state_set = self.get_state_set()
        if state_set is not None:

            # Mainly firefox elements after the workaround
            # for firefox 50.
            if role in [Atspi.Role.DOCUMENT_FRAME,
                        Atspi.Role.LINK,
                        ] \
               and state_set is not None and \
               not state_set.contains(Atspi.StateType.EDITABLE):
                    return True
        return False

    def is_single_line(self):
        """ Is accessible a single line text entry? """
        state_set = self.get_state_set()
        return state_set and state_set.contains(Atspi.StateType.SINGLE_LINE)

    def is_toolkit_gtk3(self):
        """ Are the accessible attributes from a gtk3 widget? """
        attributes = self.get_attributes()
        return attributes and \
            "toolkit" in attributes and attributes["toolkit"] == "gtk"

    def is_toolkit_qt(self):
        """
        Are the accessible attributes from a Qt widget?

        Two sources are consulted in order:

        1) The widget-level "toolkit" attribute. GTK widgets expose
           toolkit info this way; Qt widgets on KDE Plasma typically
           don't.

        2) The application accessible's get_toolkit_name(). AT-SPI's
           canonical location for toolkit metadata is the application
           root, not individual widgets. Qt apps (Kate / KWrite /
           Konsole / Dolphin / etc.) report "Qt" here while their
           widgets' own get_toolkit_name() returns None.

        Required on Wayland: the uinput keysynth path injects raw
        keycodes that the compositor interprets through the currently-
        active xkb layout group, so cross-group keysyms (e.g. typing
        '>' or '<' or Latin snippets while a non-Latin group is
        active) come out wrong. Routing through AT-SPI direct
        insertion bypasses that entirely. Qt 5.15+ / Qt 6 on modern
        Linux implement EditableText.insert correctly.
        """
        # (1) widget-level "toolkit" attribute (where GTK reports "gtk")
        attributes = self.get_attributes()
        if attributes:
            tk = attributes.get("toolkit", "") or ""
            if tk.lower().startswith("qt"):
                return True

        # (2) application-level toolkit name (where Qt reports "Qt")
        app_tk = self.get_app_toolkit_name() or ""
        if app_tk.lower().startswith("qt"):
            return True

        return False

    def get_character_extents(self, accessible, offset):
        """ Screen rect of the character at offset """
        try:
            rect = self._get_character_extents(offset)
        except Exception as ex:  # private exception gi._glib.GError when
                # right clicking onboards unity2d launcher (Precise)
            _logger.atspi("Invalid accessible,"
                          " failed to get character extents: " +
                          unicode_str(ex))
            rect = Rect()
        return rect

    def _get_character_extents(self, offset):
        """
        Screen rect of the character at offset of the accessible, little
        caching and exception handling.
        """
        scale = config.window_scaling_factor
        if scale != 1.0:
            # Only Gtk-3 widgets return scaled coordinates, all others,
            # including Gtk-2 apps like firefox, clawsmail and Qt-apps,
            # apparently don't.
            if self.is_toolkit_gtk3():
                scale = 1.0
            else:
                scale = 1.0 / config.window_scaling_factor

        ext = self._accessible.get_character_extents(offset,
                                                     Atspi.CoordType.SCREEN)
        # x, y = ext.x + ext.width / 2, ext.y + ext.height / 2
        # offset_control = self._accessible.get_offset_at_point(x, y,
        #                                                Atspi.CoordType.SCREEN)
        # print(offset, offset_control)
        return Rect(ext.x * scale, ext.y * scale,
                    ext.width * scale, ext.height * scale)


class AsyncEvent:
    """
    Decouple AT-SPI events from D-Bus calls to reduce the risk for deadlocks.
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self._kwargs = kwargs

    def __repr__(self):
        return type(self).__name__ + "(" + \
            ", ".join(str(key) + "=" + repr(val)
                      for key, val in self._kwargs.items()) \
            + ")"


class AtspiStateTracker(EventSource):
    """
    Keeps track of the currently active accessible by listening
    to AT-SPI focus events.
    """

    _focus_event_names      = ("text-entry-activated",)
    _text_event_names       = ("text-changed", "text-caret-moved")
    _key_stroke_event_names = ("key-pressed",)
    _async_event_names      = ("async-focus-changed",
                               "async-text-changed",
                               "async-text-caret-moved")
    _event_names = (_async_event_names +
                    _focus_event_names +
                    _text_event_names +
                    _key_stroke_event_names)

    _focus_listeners_registered = False
    _keystroke_listeners_registered = False
    _text_listeners_registered = False

    _keystroke_listener = None

    # asynchronously accessible members
    _focused_accessible = None   # last focused editable accessible
    _focused_pid = None          # pid of last focused editable accessible
    _active_accessible = None    # currently active editable accessible
    _active_accessible_activation_time = 0.0  # time since focus received
    _last_active_accessible = None

    _poll_unity_timer = Timer()

    def __new__(cls, *args, **kwargs):
        """
        Singleton magic.
        """
        if not hasattr(cls, "self"):
            cls.self = object.__new__(cls, *args, **kwargs)
            cls.self.construct()
        return cls.self

    def __init__(self):
        """
        Called multiple times, don't use this.
        """
        pass

    def construct(self):
        """
        Singleton constructor, runs only once.
        """
        EventSource.__init__(self, self._event_names)

        self._frozen = False

    def cleanup(self):
        EventSource.cleanup(self)
        self._register_atspi_listeners(False)

    def connect(self, event_name, callback):
        EventSource.connect(self, event_name, callback)
        self._update_listeners()

    def disconnect(self, event_name, callback):
        had_listeners = self.has_listeners(self._event_names)

        EventSource.disconnect(self, event_name, callback)
        self._update_listeners()

        # help debugging disconnecting events on exit
        if had_listeners and not self.has_listeners(self._event_names):
            _logger.info("all listeners disconnected")

    def _update_listeners(self):
        register = self.has_listeners(self._focus_event_names)
        self._register_atspi_focus_listeners(register)

        register = self.has_listeners(self._text_event_names)
        self._register_atspi_text_listeners(register)

        register = self.has_listeners(self._key_stroke_event_names)
        self._register_atspi_keystroke_listeners(register)

    def _register_atspi_listeners(self, register):
        self._register_atspi_focus_listeners(register)
        self._register_atspi_text_listeners(register)
        self._register_atspi_keystroke_listeners(register)

    def _register_atspi_focus_listeners(self, register):
        if "Atspi" not in globals():
            return

        if self._focus_listeners_registered != register:

            if register:
                self.atspi_connect("_listener_focus",
                                   "focus",
                                   self._on_atspi_global_focus)
                self.atspi_connect("_listener_object_focus",
                                   "object:state-changed:focused",
                                   self._on_atspi_object_focus)

                # private asynchronous events
                for name in self._async_event_names:
                    handler = "_on_" + name.replace("-", "_")
                    EventSource.connect(self, name, getattr(self, handler))
            else:
                self._poll_unity_timer.stop()

                self.atspi_disconnect("_listener_focus",
                                      "focus")
                self.atspi_disconnect("_listener_object_focus",
                                      "object:state-changed:focused")

                for name in self._async_event_names:
                    handler = "_on_" + name.replace("-", "_")
                    EventSource.disconnect(self, name, getattr(self, handler))

            self._focus_listeners_registered = register

    def _register_atspi_text_listeners(self, register):
        if "Atspi" not in globals():
            return

        if self._text_listeners_registered != register:
            if register:
                self.atspi_connect("_listener_text_changed",
                                   "object:text-changed",
                                   self._on_atspi_text_changed)
                self.atspi_connect("_listener_text_caret_moved",
                                   "object:text-caret-moved",
                                   self._on_atspi_text_caret_moved)

            else:
                self.atspi_disconnect("_listener_text_changed",
                                      "object:text-changed")
                self.atspi_disconnect("_listener_text_caret_moved",
                                      "object:text-caret-moved")

        self._text_listeners_registered = register

    def _register_atspi_keystroke_listeners(self, register):
        if "Atspi" not in globals():
            return

        if self._keystroke_listeners_registered != register:
            modifier_masks = range(16)

            if register:
                if not self._keystroke_listener:
                    self._keystroke_listener = \
                        Atspi.DeviceListener.new(self._on_atspi_keystroke,
                                                 None)

                for modifier_mask in modifier_masks:
                    Atspi.register_keystroke_listener(
                        self._keystroke_listener,
                        None,        # key set, None=all
                        modifier_mask,
                        Atspi.KeyEventType.PRESSED,
                        Atspi.KeyListenerSyncType.SYNCHRONOUS)
            else:
                # Apparently any single deregister call will turn off
                # all the other registered modifier_masks too. Since
                # deregistering takes extremely long (~2.5s for 16 calls)
                # seize the opportunity and just pick a single arbitrary
                # mask (Quantal).
                modifier_masks = [2]

                for modifier_mask in modifier_masks:
                    Atspi.deregister_keystroke_listener(
                        self._keystroke_listener,
                        None,  # key set, None=all
                        modifier_mask,
                        Atspi.KeyEventType.PRESSED)

        self._keystroke_listeners_registered = register

    def atspi_connect(self, attribute, event, callback):
        """
        Start listening to an AT-SPI event.
        Creates a new event listener for each event, since this seems
        to be the only way to allow reliable deregistering of events.
        """
        if hasattr(self, attribute):
            listener = getattr(self, attribute)
        else:
            listener = None

        if listener is None:
            listener = Atspi.EventListener.new(callback, None)
            setattr(self, attribute, listener)
        listener.register(event)

    def atspi_disconnect(self, attribute, event):
        """
        Stop listening to AT-SPI event.
        """
        listener = getattr(self, attribute)
        listener.deregister(event)

    def freeze(self):
        """
        Freeze AT-SPI message processing, e.g. while displaying
        a dialog or popoup menu.
        """
        self._register_atspi_listeners(False)
        self._frozen = True

    def thaw(self):
        """
        Resume AT-SPI message processing.
        """
        self._update_listeners()
        self._frozen = False

    def emit_async(self, event_name, *args, **kwargs):
        if not self._frozen:
            EventSource.emit_async(self, event_name, *args, **kwargs)

    def _get_cached_accessible(self, accessible):
        return CachedAccessible(accessible) \
            if accessible else None

    # ######### synchronous handlers ######### #

    def _on_atspi_global_focus(self, event, user_data):
        self._on_atspi_focus(event, True)

    def _on_atspi_object_focus(self, event, user_data):
        self._on_atspi_focus(event)

    def _on_atspi_focus(self, event, focus_received=False):
        focused = (bool(focus_received) or
                   bool(event.detail1))  # received focus?
        ae = AsyncEvent(accessible=self._get_cached_accessible(event.source),
                        focused=focused)
        self.emit_async("async-focus-changed", ae)

    def _on_atspi_text_changed(self, event, user_data):
        # print("_on_atspi_text_changed", event.detail1, event.detail2,
        #       event.source, event.type, event.type.endswith("delete"))
        ae = AsyncEvent(accessible=self._get_cached_accessible(event.source),
                        type=event.type,
                        pos=event.detail1,
                        length=event.detail2)
        self.emit_async("async-text-changed", ae)
        return False

    def _on_atspi_text_caret_moved(self, event, user_data):
        # print("_on_atspi_text_caret_moved", event.detail1, event.detail2,
        #       event.source, event.type, event.source.get_name(),
        #       event.source.get_role())
        ae = AsyncEvent(accessible=self._get_cached_accessible(event.source),
                        caret=event.detail1)
        self.emit_async("async-text-caret-moved", ae)
        return False

    def _on_atspi_keystroke(self, event, user_data):
        if event.type == Atspi.EventType.KEY_PRESSED_EVENT:
            _logger.atspi("key-stroke {} {} {} {}"
                          .format(event.modifiers,
                                  event.hw_code, event.id, event.is_text))
            # keysym = event.id # What is this? Not an XK_ keysym apparently.
            ae = AsyncEvent(hw_code=event.hw_code,
                            modifiers=event.modifiers)
            self.emit_async("key-pressed", ae)

        return False  # don't consume event

    # ######### asynchronous handlers ######### #
    def _on_async_focus_changed(self, event):
        accessible = event.accessible
        focused = event.focused

        # Don't access the accessible while frozen. This leads to deadlocks
        # while displaying Onboard's own dialogs/popup menu's.
        if self._frozen:
            return

        self._log_accessible(accessible, focused)

        if not accessible:
            return

        # Ignore AT-SPI focus events that originate from Onboard's own
        # accessibility tree. When the user clicks a key, the Onboard
        # window emits focus events for its own buttons / window; if
        # processed, _handle_focus_changed_apps drops the target app's
        # editable accessible (PID mismatch heuristic), and the next
        # press_unicode falls back to the broken keysynth path
        # instead of using the AT-SPI direct-insert that was correctly
        # armed when the target was focused.
        try:
            if accessible.get_pid() == os.getpid():
                _logger.atspi(
                    "ignoring AT-SPI focus event from our own process "
                    "(pid={})".format(os.getpid()))
                return
        except Exception:
            pass

        app_name = accessible.get_app_name().lower()
        if app_name == "unity":
            self._handle_focus_changed_unity(event)
        else:
            self._handle_focus_changed_apps(event)

    def _handle_focus_changed_apps(self, event):
        """ Focus change in regular applications """
        accessible = event.accessible
        focused = event.focused

        # Since Trusty, focus events no longer come reliably in a
        # predictable order. -> Store the last editable accessible
        # so we can pick it over later focused non-editable ones.
        # Helps to keep the keyboard open in presence of popup selections
        # e.g. in GNOME's file dialog and in Unity Dash.
        if self._focused_accessible == accessible:
            if not focused:
                self._focused_accessible = None
        else:
            pid = accessible.get_pid()

            if focused:
                self._poll_unity_timer.stop()

                if accessible.is_editable():
                    self._focused_accessible = accessible
                    self._focused_pid = pid

                # Static accessible, i.e. something that cannot
                # accidentally steal the focus from an editable
                # accessible. e.g. firefox ATSPI_ROLE_DOCUMENT_FRAME?
                elif accessible.is_not_focus_stealing():
                    self._focused_accessible = None
                    self._focused_pid = None

                else:
                    # Wily: attempt to hide when unity dash closes
                    # (there's no focus lost event).
                    # Also check duration since last activation to
                    # skip out of order focus events (firefox
                    # ATSPI_ROLE_DOCUMENT_FRAME) for a short while
                    # after opening dash.
                    now = time.time()
                    if focused and \
                       now - self._active_accessible_activation_time > .5:
                        if self._focused_pid != pid:
                            self._focused_accessible = None
                            _logger.atspi("Dropping accessible due to "
                                          "pid change: {} != {} "
                                          .format(self._focused_pid, pid))

        # Has the previously focused accessible lost the focus?
        active_accessible = self._focused_accessible
        if active_accessible and \
           not active_accessible.is_focused(True):

            # Zesty: Firefox 50+ loses focus of the URL entry after
            # typing just a few letters and focuses a completion
            # menu item instead. Let's pretend the accessible is
            # still focused in that case.
            is_firefox_completion = \
                self._focused_accessible.is_urlbar() and \
                accessible.get_role() == Atspi.Role.MENU_ITEM

            if not is_firefox_completion:
                active_accessible = None

        self._set_active_accessible(active_accessible)

    def _handle_focus_changed_unity(self, event):
        """ Focus change in Unity Dash """
        accessible = event.accessible
        focused = event.focused

        # Wily: prevent random icons, buttons and toolbars
        # in unity dash from hiding Onboard. Somehow hovering
        # over those buttons silently drops the focus from the
        # text entry. Let's pretend the buttons don't exist
        # and keep the previously saved text entry active.

        # Zesty: Don't fight lost focus events anymore, only
        # react to focus events when the text entry gains focus.
        if focused and \
           accessible.is_editable():
            self._focused_accessible = accessible
            self._set_active_accessible(accessible)

            # For hiding we poll Dash's toplevel accessible
            def _poll_unity_dash():
                frame = accessible.get_frame()
                state_set = frame.get_state_set()

                _logger.debug(
                    "polling unity dash state_set: {}"
                    .format(AtspiStateType.to_strings(state_set)))

                if not state_set or \
                   not state_set.contains(Atspi.StateType.ACTIVE):
                    self._focused_accessible = None
                    self._set_active_accessible(None)
                    return False

                return True

            # Only ever start polling if Dash is "ACTIVE".
            # The state_set might change in the future and the
            # keyboard better fail to auto-hide than to never show.
            frame = accessible.get_frame()
            state_set = frame.get_state_set()

            _logger.debug(
                "dash focused, state_set: {}"
                .format(AtspiStateType.to_strings(state_set)))

            if state_set and \
               state_set.contains(Atspi.StateType.ACTIVE):
                self._poll_unity_timer.start(0.5, _poll_unity_dash)

    def _set_active_accessible(self, accessible):
        if self._active_accessible != accessible:
            self._active_accessible = accessible

            if self._active_accessible or \
               self._last_active_accessible:

                # notify listeners
                self.emit("text-entry-activated", self._active_accessible)

                self._last_active_accessible = self._active_accessible
                self._active_accessible_activation_time = time.time()

    def _on_async_text_changed(self, event):
        if event.accessible == self._active_accessible:

            type = event.type
            insert = type.endswith(("insert", "insert:system"))
            delete = type.endswith(("delete", "delete:system"))
            # print(event.accessible.get_id(), type, insert)

            _logger.atspi("_on_async_text_changed: "
                          "type={} insert={} delete={}"
                          .format(type, insert, delete))

            if insert or delete:
                event.insert = insert
                self.emit("text-changed", event)
            else:
                _logger.warning("_on_async_text_changed: "
                                "unknown event type '{}'"
                                .format(event.type))

    def _on_async_text_caret_moved(self, event):
        if event.accessible == self._active_accessible:
            self.emit("text-caret-moved", event)

    def _log_accessible(self, accessible, focused):
        if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
            msg = "AT-SPI focus event: focused={}, ".format(focused)
            msg += "accessible={}, ".format(accessible)

            if accessible:
                name = accessible.get_name()
                role = accessible.get_role()
                role_name = accessible.get_role_name()
                state_set = accessible.get_state_set()
                states = state_set.states
                editable = state_set.contains(Atspi.StateType.EDITABLE) \
                    if state_set else None
                extents = accessible.get_extents()

                msg += "name={name}, role={role}({role_name}), " \
                       "editable={editable}, states={states}, " \
                       "extents={extents}]" \
                       .format(accessible=accessible, name=repr(name),
                               role=role.value_name if role else role,
                               role_name=repr(role_name),
                               editable=editable,
                               states=states,
                               extents=extents
                               )
            _logger.atspi(msg)


class AtspiStateType:
    states = ['ACTIVE',
              'ANIMATED',
              'ARMED',
              'BUSY',
              'CHECKED',
              'COLLAPSED',
              'DEFUNCT',
              'EDITABLE',
              'ENABLED',
              'EXPANDABLE',
              'EXPANDED',
              'FOCUSABLE',
              'FOCUSED',
              'HAS_TOOLTIP',
              'HORIZONTAL',
              'ICONIFIED',
              'INDETERMINATE',
              'INVALID',
              'INVALID_ENTRY',
              'IS_DEFAULT',
              'LAST_DEFINED',
              'MANAGES_DESCENDANTS',
              'MODAL',
              'MULTISELECTABLE',
              'MULTI_LINE',
              'OPAQUE',
              'PRESSED',
              'REQUIRED',
              'RESIZABLE',
              'SELECTABLE',
              'SELECTABLE_TEXT',
              'SELECTED',
              'SENSITIVE',
              'SHOWING',
              'SINGLE_LINE',
              'STALE',
              'SUPPORTS_AUTOCOMPLETION',
              'TRANSIENT',
              'TRUNCATED',
              'VERTICAL',
              'VISIBLE',
              'VISITED',
              ]

    @staticmethod
    def to_strings(state_set):
        result = []
        if state_set is not None:
            for s in AtspiStateType.states:
                if state_set.contains(getattr(Atspi.StateType, s)):
                    result.append(s)
        return result





# -*- coding: UTF-8 -*-

# Copyright © 2007 Martin Böhme <martin.bohm@kubuntu.org>
# Copyright © 2009 Chris Jones <tortoise@tortuga>
# Copyright © 2010 Francesco Fumanti <francesco.fumanti@gmx.net>
# Copyright © 2011 Alan Bell <alanbell@ubuntu.com>
# Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
# Copyright © 2009-2014, 2016 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function, unicode_literals

from math import pi, sin, cos, sqrt

import cairo
from Onboard.Version import require_gi_versions
require_gi_versions()
from gi.repository import GLib, Gdk, Pango, PangoCairo, GdkPixbuf

from Onboard.KeyCommon   import (KeyCommon, RectKeyCommon,
                                 LOD, InputlineKeyCommon, ImageStyle,
                                 ImageSlot)
from Onboard.KeyCommon   import *
from Onboard.WindowUtils import DwellProgress
from Onboard.utils       import (brighten, unicode_str,
                                 gradient_line, drop_shadow,
                                 roundrect_curve, roundrect_curve_custom,
                                 rounded_path,
                                 rounded_polygon_path_to_cairo_path)

import logging
_logger = logging.getLogger("KeyGTK")

from Onboard.Config import Config
config = Config()

PangoUnscale = 1.0 / Pango.SCALE


class Key(KeyCommon):
    _pango_layouts = None
    _label_extents = None  # resolution independent size {mod_mask: (w, h)}
    _popup_indicator = ""  # font dependent popup indicator (ellipsis)

    _shadow_steps  = 0
    _shadow_alpha  = 0
    _shadow_presets = ((1, 0.015), (4, 0.005)) # quality presets (steps, alpha)

    def __init__(self):
        KeyCommon.__init__(self)
        self._label_extents = {}

    def get_best_font_size(self):
        """
        Get the maximum font possible that would not cause the label to
        overflow the boundaries of the key.
        """

        raise NotImplementedError()

    @staticmethod
    def reset_pango_layout():
        Key._pango_layouts = None

    @staticmethod
    def get_pango_layout(text, font_size, slot = 0):
        # work around memory leak (gnome #599730)
        if Key._pango_layouts is None:
            # use PangoCairo.create_layout once it works with gi (pango >= 1.29.1)
            #Key._pango_layouts = PangoCairo.create_layout(context)
            Key._pango_layouts = (
                Pango.Layout(context = Gdk.pango_context_get()),
                Pango.Layout(context = Gdk.pango_context_get()),
                Pango.Layout(context = Gdk.pango_context_get()))

        layout = Key._pango_layouts[slot]
        Key.prepare_pango_layout(layout, text, font_size)
        return layout

    @staticmethod
    def prepare_pango_layout(layout, text, font_size):
        if text is None:
            text = ""
        layout.set_text(text, -1)
        layout.set_width(-1) # no wrapping, ellipsization
        font_description = Pango.FontDescription(config.theme_settings.key_label_font)
        font_description.set_size(max(1, font_size))
        layout.set_font_description(font_description)

    @classmethod
    def set_shadow_quality(_class, quality):
        if quality is None:
            quality = 1
        _class._shadow_steps, _class._shadow_alpha = \
                                    _class._shadow_presets[quality]


class RectKey(Key, RectKeyCommon, DwellProgress):

    _image_pixbuf = None
    _requested_image_size = None
    _shadow_surface = None

    can_draw_cached = True

    def __init__(self, id="", border_rect=None):
        Key.__init__(self)
        RectKeyCommon.__init__(self, id, border_rect)

        self._key_surfaces = {}

    def is_key(self):
        """ Is this a key item? """
        return True

    def invalidate_caches(self):
        """
        Clear buffered patterns, e.g. after resizing, change of settings...
        """
        self.invalidate_key()
        self.invalidate_shadow()

    def invalidate_key(self):
        self._key_surfaces = {}

    def invalidate_image(self):
        """
        Images only have to be expicitely cleared when the
        window_scaling_factor changes.
        """
        self._image_pixbuf = {}
        self._requested_image_size = {}

    def invalidate_shadow(self):
        self._shadow_surface = None

    def set_border_rect(self, rect):
        """
        The expand-corrections button moves around a lot.
        Be sure to keep its image surfaces updated.
        """
        if rect != self.get_border_rect():
            super(RectKey, self).set_border_rect(rect)
            self.invalidate_caches()

    def draw_cached(self, cr):
        key = (self.label, self.font_size >> 8)
        entry = self._key_surfaces.get(key)
        if entry is None:
            if self.font_size:
                entry = self._create_key_surface(cr)
                self._key_surfaces[key] = entry

        if entry:
            surface, rect = entry
            cr.set_source_surface(surface, rect.x, rect.y)
            cr.paint()

    def _create_key_surface(self, base_context):
        rect = self.get_canvas_rect()
        clip_rect = rect.inflate(*self.get_extra_render_size()).int()

        # create caching surface
        target = base_context.get_target()
        surface = target.create_similar(cairo.CONTENT_COLOR_ALPHA,
                                        clip_rect.w, clip_rect.h)
        cr = cairo.Context(surface)

        cr.save()
        cr.translate(-clip_rect.x, -clip_rect.y)
        self.draw(cr)
        cr.restore()

        Gdk.flush()  # else artefacts in labels and images on Nexus 7, Raring

        return surface, clip_rect

    def draw_item(self, context):
        if context.draw_cached and self.can_draw_cached:
            self.draw_cached(context.cr)
        else:
            self.draw(context.cr, context.lod)

    def draw(self, cr, lod=LOD.FULL):
        self.draw_geometry(cr, lod)
        self.draw_image(cr, lod)
        self.draw_label(cr, lod)

    def draw_geometry(self, cr, lod):
        if not self.show_face and not self.show_border:
            return

        if lod == LOD.FULL and self.show_border:
            scale = self.get_stroke_width()
            if scale:
                root = self.get_layout_root()
                t    = root.context.scale_log_to_canvas((1.0, 1.0))
                line_width = (t[0] + t[1]) / 2.4
                line_width = min(line_width, 3.0) * scale
                line_width = max(line_width, 1.0)
            else:
                line_width = 0
        else:
            line_width = 0

        fill = self.get_fill_color()

        key_style = self.get_style()
        if key_style == "flat":
            self.draw_flat_key(cr, fill, line_width)

        elif key_style == "gradient":
            self.draw_gradient_key(cr, fill, line_width, lod)

        elif key_style == "dish":
            self.draw_dish_key(cr, fill, line_width, lod)

    def draw_flat_key(self, cr, fill, line_width):
        self._build_canvas_path(cr)

        if self.show_face:
            cr.set_source_rgba(*fill)
            if line_width:
                cr.fill_preserve()
            else:
                cr.fill()

        if line_width:
            cr.set_source_rgba(*self.get_stroke_color())
            cr.set_line_width(line_width)
            cr.stroke()

    def draw_gradient_key(self, cr, fill, line_width, lod):
        # simple gradients for fill and stroke
        fill_gradient   = config.theme_settings.key_fill_gradient / 100.0
        stroke_gradient = self.get_stroke_gradient()
        alpha = self.get_gradient_angle()

        rect = self.get_canvas_rect()
        self._build_canvas_path(cr, rect)
        gline = gradient_line(rect, alpha)

        # fill
        if self.show_face:
            if fill_gradient and lod:
                pat = cairo.LinearGradient (*gline)
                rgba = brighten(+fill_gradient*.5, *fill)
                pat.add_color_stop_rgba(0, *rgba)
                rgba = brighten(-fill_gradient*.5, *fill)
                pat.add_color_stop_rgba(1, *rgba)
                cr.set_source (pat)
            else: # take gradient from color scheme (not implemented)
                cr.set_source_rgba(*fill)

            if self.show_border:
                cr.fill_preserve()
            else:
                cr.fill()

        # stroke
        if self.show_border:
            if stroke_gradient:
                if lod:
                    stroke = fill
                    pat = cairo.LinearGradient(*gline)

                    if self.pressed:
                        stops = [-stroke_gradient * 0.5, +stroke_gradient * 0.5]
                    else:
                        stops = [+stroke_gradient * 0.5, -stroke_gradient * 0.5]

                    for i, stop in enumerate(stops):
                        rgba = brighten(stop, *stroke)
                        pat.add_color_stop_rgba(i, *rgba)

                    cr.set_source(pat)
                else:
                    cr.set_source_rgba(*fill)
            else:
                cr.set_source_rgba(*self.get_stroke_color())

            cr.set_line_width(line_width)
        cr.stroke()

    def draw_dish_key(self, cr, fill, line_width, lod):
        canvas_rect = self.get_canvas_rect()
        if self.geometry:
            geometry = self.geometry
        else:
            geometry = KeyGeometry.from_rect(self.get_border_rect())
        size_scale_x, size_scale_y = geometry.scale_log_to_size((1.0, 1.0))

        # compensate for smaller size due to missing stroke
        canvas_rect = canvas_rect.inflate(1.0)

        # parameters for the base path
        base_rgba = brighten(-0.200, *fill)
        stroke_gradient = self.get_stroke_gradient()
        light_dir = self.get_light_direction() - pi * 0.5  # 0 = light from top
        lightx = cos(light_dir)
        lighty = sin(light_dir)

        key_offset_x, key_offset_y, key_size_x, key_size_y = \
                                            self.get_key_offset_size(geometry)
        radius_pct = max(config.theme_settings.roundrect_radius, 2)
        radius_pct = max(radius_pct, 2) # too much +-1 fudging for square corners
        chamfer_size = self.get_chamfer_size()
        chamfer_size = (self.context.scale_log_to_canvas_x(chamfer_size) +
                        self.context.scale_log_to_canvas_y(chamfer_size)) * 0.5

        # parameters for the top path, key face
        stroke_width  = self.get_stroke_width()
        key_offset_top_y = key_offset_y - \
            config.DISH_KEY_Y_OFFSET * stroke_width
        border = config.DISH_KEY_BORDER
        scale_top_x = 1.0 - (border[0] * stroke_width * size_scale_x * 2.0)
        scale_top_y = 1.0 - (border[1] * stroke_width * size_scale_y * 2.0)
        key_size_top_x = key_size_x * scale_top_x
        key_size_top_y = key_size_y * scale_top_y
        chamfer_size_top = chamfer_size * (scale_top_x + scale_top_y) * 0.5

        # realize all paths we're going to use
        polygons, polygon_paths = \
            self.get_canvas_polygons(geometry,
                                   key_offset_x, key_offset_y,
                                   key_size_x, key_size_y,
                                   radius_pct, chamfer_size)
        polygons_top, polygon_paths_top = \
            self.get_canvas_polygons(geometry,
                                   key_offset_x, key_offset_top_y,
                                   key_size_top_x - size_scale_x,
                                   key_size_top_y - size_scale_y,
                                   radius_pct, chamfer_size_top)
        polygons_top1, polygon_paths_top1 = \
            self.get_canvas_polygons(geometry,
                                   key_offset_x, key_offset_top_y,
                                   key_size_top_x, key_size_top_y,
                                   radius_pct, chamfer_size_top)

        # draw key border
        if self.show_border:
            if not lod:
                cr.set_source_rgba(*base_rgba)
                for path in polygon_paths:
                    rounded_polygon_path_to_cairo_path(cr, path)
                    cr.fill()
            else:
                for ipg, polygon in enumerate(polygons):
                    polygon_top = polygons_top[ipg]
                    path = polygon_paths[ipg]
                    path_top = polygon_paths_top[ipg]

                    self._draw_dish_key_border(cr, path, path_top,
                                               polygon, polygon_top,
                                               base_rgba, stroke_gradient,
                                               lightx, lighty)

        # Draw the key face, the smaller top rectangle.
        if self.show_face:
            if not lod:
                cr.set_source_rgba(*fill)
            else:
                # Simulate the concave key dish with a gradient that has
                # a sligthly brighter middle section.
                if self.id == "SPCE":
                    angle = pi / 2.0  # space has a convex top
                else:
                    angle = 0.0       # all others are concave
                fill_gradient   = config.theme_settings.key_fill_gradient / 100.0
                dark_rgba = brighten(-fill_gradient*.5, *fill)
                bright_rgba = brighten(+fill_gradient*.5, *fill)
                gline = gradient_line(canvas_rect, angle)

                pat = cairo.LinearGradient (*gline)
                pat.add_color_stop_rgba(0.0, *dark_rgba)
                pat.add_color_stop_rgba(0.5, *bright_rgba)
                pat.add_color_stop_rgba(1.0, *dark_rgba)
                cr.set_source (pat)

            for path in polygon_paths_top1:
                rounded_polygon_path_to_cairo_path(cr, path)
                cr.fill()

    def _draw_dish_key_border(self, cr, path, path_top,
                              polygon, polygon_top,
                              base_rgba, stroke_gradient, lightx, lighty):
        n = len(polygon)
        m = len(path)

        # Lambert lighting
        edge_colors = []
        for i in range(0, n, 2):
            x0 = polygon[i]
            y0 = polygon[i+1]
            if i < n-2:
                x1 = polygon[i+2]
                y1 = polygon[i+3]
            else:
                x1 = polygon[0]
                y1 = polygon[1]

            nx = y1 - y0
            ny = -(x1 - x0)
            ln = sqrt(nx*nx + ny*ny)
            I = (nx * lightx + ny * lighty) / ln \
                * stroke_gradient * 0.8 \
                if ln else 0.0
            edge_colors.append(brighten(I, *base_rgba))

        # draw border sections
        edge = 0
        for i in range(0, m-2, 2):
            # get path points
            i1 = i + 1
            i2 = i + 2
            if i2 >= m:
                i2 -= m
            i3 = i + 3
            if i3 >= m:
                i3 = 1

            p0 = path[i]
            p0x = p0[0]
            p0y = p0[1]

            p1 = path[i1]
            p1x = p1[0]
            p1y = p1[1]

            p2 = path[i2]
            p2x = p2[0]
            p2y = p2[1]

            p3 = path[i3]
            p3x = p3[0]
            p3y = p3[1]

            p = path_top[i]
            ptop0x = p[0]
            ptop0y = p[1]

            p = path_top[i1]
            ptop1x = p[0]
            ptop1y = p[1]

            p = path_top[i2]
            ptop2x = p[0]
            ptop2y = p[1]

            # get polygon points, only to
            # fill in gaps at concave corners.
            j0 = edge*2
            j1 = j0 + 2
            if j1 >= n:
                j1 -= n
            j2 = j0 + 4
            if j2 >= n:
                j2 -= n

            ptopax = polygon_top[j0]
            ptopay = polygon_top[j0 + 1]
            ptopbx = polygon_top[j1]
            ptopby = polygon_top[j1 + 1]
            ptopcx = polygon_top[j2]
            ptopcy = polygon_top[j2 + 1]
            vax = ptopbx - ptopax
            vay = ptopby - ptopay
            nbx = ptopcy - ptopby
            nby = -(ptopcx - ptopbx)
            concave = vax*nbx + vay*nby < 0.0

            # Fake Gouraud shading: draw a gradient between mid points
            # of the lines connecting the base with the top path.
            pat = cairo.LinearGradient((p1x + ptop1x) * 0.5,
                                        (p1y + ptop1y) * 0.5,
                                        (p2x + ptop2x) * 0.5,
                                        (p2y + ptop2y) * 0.5)
            edge1 = (edge + 1) % len(edge_colors)
            pat.add_color_stop_rgba(0.0, *edge_colors[edge])
            pat.add_color_stop_rgba(1.0, *edge_colors[edge1])
            cr.set_source (pat)

            # Draw corners and edges with enough overlap to avoid
            # artefacts at touching line boundaries.
            cr.move_to(p0x, p0y)
            cr.line_to(p1x, p1y)
            cr.curve_to(p2[2], p2[3], p2[4], p2[5], p2[0], p2[1])
            cr.line_to(p3x, p3y)
            cr.line_to(ptop2x, ptop2y)
            if concave:
                cr.line_to(ptopbx, ptopby)
            cr.line_to(ptop1x, ptop1y)
            cr.line_to(ptop0x, ptop0y)
            cr.close_path()
            cr.fill()

            edge += 1

    def get_label_runs(self):
        runs = []
        log_rect = self.get_label_rect()
        canvas_rect = self.context.log_to_canvas_rect(log_rect)

        # secondary label
        label = self.get_secondary_label()
        if label and \
           len(label) == 1 and \
           config.keyboard.show_secondary_labels:
            font_size = self.font_size * 0.5
            layout = self.get_pango_layout(label, font_size, 1)
            src_size = layout.get_size()
            src_size = (src_size[0] * PangoUnscale, src_size[1] * PangoUnscale)
            xalign, yalign = self.align_secondary_label(src_size,
                                                (canvas_rect.w, canvas_rect.h))
            x = int(canvas_rect.x + xalign)
            y = int(canvas_rect.y + yalign)
            rgba = self.get_secondary_label_color()

            runs.append((layout, x, y, rgba))

        # popup indicator
        if not self.popup_id is None and \
           not config.xid_mode:

            label = self._get_popup_indicator()
            font_size = self.font_size
            layout = self.get_pango_layout(label, font_size, 2)

            src_size = layout.get_size()
            src_size = (src_size[0] * PangoUnscale, src_size[1] * PangoUnscale)
            xalign, yalign = self.align_popup_indicator(src_size,
                                                 (canvas_rect.w, canvas_rect.h))
            x = int(canvas_rect.x + xalign)
            y = int(canvas_rect.y + yalign)
            rgba = self.get_secondary_label_color()

            runs.append((layout, x, y, rgba))

        # main label
        label = self.get_label()
        if label:
            font_size = self.font_size
            layout = self.get_pango_layout(label, font_size, 0)
            src_size = layout.get_size()
            src_size = (src_size[0] * PangoUnscale, src_size[1] * PangoUnscale)
            xalign, yalign = self.align_label(src_size,
                                                (canvas_rect.w, canvas_rect.h))
            x = int(canvas_rect.x + xalign)
            y = int(canvas_rect.y + yalign)
            rgba = self.get_label_color()

            runs.append((layout, x, y, rgba))

        return runs

    def _get_popup_indicator(self):
        """
        Find the shortest ellipsis possible with the current font.
        The font is assumed to never change during the livetime of the key.
        """
        result = self._popup_indicator
        if not result:
            labels = ("…", "...")  # label candidates

            BASE_FONTDESCRIPTION_SIZE = 10000000
            wmin = None
            result = ""
            for label in labels:
                layout = self.get_pango_layout(label, BASE_FONTDESCRIPTION_SIZE, 2)
                w = layout.get_size()[0]
                if wmin is None or w < wmin:
                    wmin = w
                    result = label
            self._popup_indicator = result

        return result

    def draw_label(self, context, lod):
        # Skip cairo errors when drawing labels with font size 0
        # This may happen for hidden keys and keys with bad size groups.
        if self.font_size == 0 or not self.show_label:
            return

        runs = self.get_label_runs()
        if not runs:
            return

        fill = self.get_fill_color()

        for dx, dy, lum, last in self._label_iterations(lod):
            # draw dwell progress after fake emboss, before final image
            if last and self.is_dwelling():
                DwellProgress.draw(self, context,
                                   self.get_dwell_progress_canvas_rect(),
                                   self.get_dwell_progress_color())
            for layout, x, y, rgba in runs:
                if lum:
                    rgba = brighten(lum, *fill) # darker
                context.move_to(x + dx, y + dy)
                context.set_source_rgba(*rgba)
                PangoCairo.show_layout(context, layout)

    def draw_image(self, context, lod):
        """
        Draws the key's optional image.
        Fixme: merge with draw_label, can't do this for 0.99 because
        the Gdk.flush() workaround on the nexus 7 might fail.
        """
        if not self.image_filenames or not self.show_image:
            return

        log_rect = self.get_label_rect()
        rect = self.context.log_to_canvas_rect(log_rect)
        if rect.w < 1 or rect.h < 1:
            return

        pixbuf = self.get_image(rect.w, rect.h)
        if not pixbuf:
            return

        src_size = (pixbuf.get_width(), pixbuf.get_height())
        xalign, yalign = self.align_label(src_size, (rect.w, rect.h))

        image_rgba = self.get_image_color()
        fill = self.get_fill_color()

        for dx, dy, lum, last in self._label_iterations(lod):
            # draw dwell progress after fake emboss, before final image
            if last and self.is_dwelling():
                DwellProgress.draw(self, context,
                                   self.get_dwell_progress_canvas_rect(),
                                   self.get_dwell_progress_color())

            r = rect.offset(xalign + dx, yalign + dy)

            if image_rgba is None:
                pixbuf.draw(context, r, None, self.image_style)
            else:
                if lum:
                    rgba = brighten(lum, *fill)  # darker
                else:
                    rgba = image_rgba

                pixbuf.draw(context, r, rgba, self.image_style)

    def draw_shadow_cached(self, context):
        entry = self._shadow_surface
        if entry is None:
            if config.theme_settings.key_shadow_strength:
                entry = self.create_shadow_surface(context,
                                              self._shadow_steps,
                                              self._shadow_alpha)
                self._shadow_surface = entry

        if entry:
            surface, rect = entry
            context.set_source_rgba(0.0, 0.0, 0.0, 1.0)
            context.mask_surface(surface, rect.x, rect.y)

    def create_shadow_surface(self, base_context, shadow_steps, shadow_alpha):
        """
        Draw shadow and shaded halo.
        Somewhat slow, make sure to cache the result.
        Glitchy, if the clip-rect covers only a single button (Precise),
        therefore, draw only with unrestricted clipping rect.
        """
        rect = self.get_canvas_rect()
        root = self.get_layout_root()

        if rect.is_empty():
            return None

        extent = min(root.context.scale_log_to_canvas((1.0, 1.0)))
        alpha = pi / 2 + self.get_light_direction()

        shadow_opacity = config.theme_settings.key_shadow_strength * \
                         shadow_alpha
        shadow_scale   = config.theme_settings.key_shadow_size / 20.0
        shadow_radius  = max(extent * shadow_scale, 1.0)
        shadow_displacement = max(extent * shadow_scale * 0.26, 1.0)
        shadow_offset  = (shadow_displacement * cos(alpha),
                          shadow_displacement * sin(alpha))

        has_halo = shadow_steps > 1 and not config.window.transparent_background
        halo_opacity   = shadow_opacity * 0.11
        halo_radius    = max(extent * 8.0, 1.0)

        clip_rect = rect.offset(shadow_offset[0]+1, shadow_offset[1]+1)
        if has_halo:
            clip_rect = clip_rect.inflate(halo_radius * 1.5)
        else:
            clip_rect = clip_rect.inflate(shadow_radius * 1.3)
        clip_rect = clip_rect.int()

        # create caching surface
        target = base_context.get_target()
        surface = target.create_similar(cairo.CONTENT_ALPHA,
                                        clip_rect.w, clip_rect.h)
        context = cairo.Context(surface)

        # paint the surface
        context.save()
        context.translate(-clip_rect.x, -clip_rect.y)

        context.rectangle(*clip_rect)
        context.clip()

        context.push_group_with_content(cairo.CONTENT_ALPHA)
        self._build_canvas_path(context, rect)
        context.set_source_rgba(0.0, 0.0, 0.0, 1.0)
        context.fill()
        shape = context.pop_group()

        # shadow
        drop_shadow(context, shape, rect,
                    shadow_radius, shadow_offset, shadow_opacity, shadow_steps)
        # halo
        if has_halo:
            drop_shadow(context, shape, rect,
                        halo_radius, shadow_offset, halo_opacity, shadow_steps)

        # cut out the key area, the key may be transparent
        context.set_operator(cairo.OPERATOR_CLEAR)
        context.set_source_rgba(0.0, 0.0, 0.0, 1.0)
        self._build_canvas_path(context, rect)
        context.fill()

        context.restore()

        return surface, clip_rect

    def _build_canvas_path(self, cr, rect = None, path = None):
        """ Build cairo path of the key geometry. """
        if self.geometry:
            if not path:
                path = self.get_canvas_path()
            self._build_complex_path(cr, path)
        else:
            if not rect:
                rect = self.get_canvas_rect()
            self.build_rect_path(cr, rect)

    def _build_complex_path(self, cr, path):
        roundness = config.theme_settings.roundrect_radius
        chamfer_size = self.get_chamfer_size()
        chamfer_size = self.context.scale_log_to_canvas_y(chamfer_size)
        rounded_path(cr, path, roundness, chamfer_size)

    def build_rect_path(self, context, rect):
        roundness = config.theme_settings.roundrect_radius
        if roundness:
            roundrect_curve(context, rect, roundness)
        else:
            context.rectangle(*rect)

    def build_rect_path_custom(self, context, rect, corner_mask):
        roundness = config.theme_settings.roundrect_radius
        if roundness:
            roundrect_curve_custom(context, rect, roundness, corner_mask)
        else:
            context.rectangle(*rect)

    def get_gradient_angle(self):
        return -pi/2.0 + self.get_light_direction()

    def get_best_font_size(self, mod_mask):
        """
        Get the maximum font size that would not cause the label to
        overflow the boundaries of the key.
        """
        # Base this on the unpressed rect, so fake physical key action
        # doesn't influence the font_size and doesn't cause surface cache
        # misses for that minor wiggle.
        rect = self.get_label_rect(self.get_unpressed_rect())
        label_width, label_height = \
                      self.get_label_base_extents(mod_mask)

        size_for_maximum_width  = self.context.scale_log_to_canvas_x(
                                      (rect.w - self.label_margin[0]*2)) \
                                  / label_width

        size_for_maximum_height = self.context.scale_log_to_canvas_y(
                                     (rect.h - self.label_margin[1]*2)) \
                                  / label_height

        if size_for_maximum_width < size_for_maximum_height:
            return int(size_for_maximum_width)
        else:
            return int(size_for_maximum_height)

    def get_label_base_extents(self, mod_mask):
        """
        Update resolution independent extents of the label layout.
        """
        extents = self._label_extents.get(mod_mask)
        if not extents:
            extents = self.calc_label_base_extents(self.get_label())
            self._label_extents[mod_mask] = extents

        return extents

    def calc_label_base_extents(self, label):
        """ Calculate font-size independent extents. """
        cr = Gdk.pango_context_get()
        layout = Pango.Layout(cr)
        BASE_FONTDESCRIPTION_SIZE = 10000000
        self.prepare_pango_layout(layout, label, BASE_FONTDESCRIPTION_SIZE)
        w, h = layout.get_size()   # In Pango units
        w = w or 1.0
        h = h or 1.0
        return w / (Pango.SCALE * BASE_FONTDESCRIPTION_SIZE), \
               h / (Pango.SCALE * BASE_FONTDESCRIPTION_SIZE)

    def invalidate_label_extents(self):
        """
        Cached label extents are resolution independent. Calling this
        is only necessary when the system font dpi change.
        """
        self._label_extents = {}

    def get_image(self, width, height):
        """
        Get the cached image pixbuf object, load image
        and create it if necessary.
        Width and height in canvas coordinates.
        """
        if not self.image_filenames:
            return None

        if self.active and ImageSlot.ACTIVE in self.image_filenames:
            slot = ImageSlot.ACTIVE
        else:
            slot = ImageSlot.NORMAL
        image_filename = self.image_filenames.get(slot)
        if not image_filename:
            return

        if not self._image_pixbuf:
            self._image_pixbuf = {}
            self._requested_image_size = {}

        pixbuf = self._image_pixbuf.get(slot)
        size = self._requested_image_size.get(slot)

        if not pixbuf or \
           size[0] != int(width) or size[1] != int(height):
            pixbuf = None
            filename = config.get_image_filename(image_filename)
            if filename:
                _logger.debug("loading image '{}'".format(filename))

                try:
                    pixbuf = PixBufScaled. \
                        from_file_and_size(filename, width, height)
                except Exception as ex: # private exception gi._glib.GError when
                                        # librsvg2-common wasn't installed
                    _logger.error("get_image(): " + unicode_str(ex))

                if pixbuf:
                    self._requested_image_size[slot] = (int(width), int(height))

            self._image_pixbuf[slot] = pixbuf

        return pixbuf

    def _label_iterations(self, lod):
        stroke_gradient = self.get_stroke_gradient()
        if lod == LOD.FULL and \
           self.get_style() != "flat" and stroke_gradient:
            root = self.get_layout_root()
            d = 0.4  # fake-emboss distance
            #d = max(src_size[1] * 0.02, 0.0)
            max_offset = 2

            alpha = self.get_gradient_angle()
            xo = root.context.scale_log_to_canvas_x(d * cos(alpha))
            yo = root.context.scale_log_to_canvas_y(d * sin(alpha))
            xo = min(int(round(xo)), max_offset)
            yo = min(int(round(yo)), max_offset)

            luminosity_factor = stroke_gradient * 0.25

            # shadow
            yield xo, yo, -luminosity_factor, False

            # highlight
            yield -xo, -yo, luminosity_factor, False

        # normal
        yield 0, 0, 0, True


class FixedFontMixin:
    """ Font size independent of text length """

    def get_best_font_size(self, mod_mask):
        """
        Get the maximum font size that would not cause the label to
        overflow the height of the key.
        """
        return self.calc_font_size(self.context,
                                   self.get_fullsize_rect().get_size(),
                                   True)

    def calc_font_size(self, context, size, use_width = False):
        """ Calculate font size based on the height of the key """
        # Base this on the unpressed rect, so fake physical key action
        # doesn't influence the font_size and doesn't cause surface cache
        # misses for that minor wiggle.
        label_width, label_height = self.get_label_base_extents(0)

        size_for_maximum_width  = context.scale_log_to_canvas_x(
                                     (size[0] - self.label_margin[0]*2)) \
                                  / label_width

        size_for_maximum_height = context.scale_log_to_canvas_y(
                                     (size[1] - self.label_margin[1]*2)) \
                                 / label_height

        font_size = size_for_maximum_height
        if use_width and size_for_maximum_width < font_size:
            font_size = size_for_maximum_width

        return int(font_size * 0.9)

    def get_label_base_extents(self, mod_mask):
        """
        Update resolution independent extents of the label layout.
        """
        extents = self._label_extents.get(mod_mask)
        if not extents:
            extents = self.calc_label_base_extents("Mg")
            self._label_extents[mod_mask] = extents

        return extents


class WordlistKey(RectKey):

    def get_style(self):
        style = super(WordlistKey, self).get_style()
        if style == "dish":
            style = "gradient"
        return style

    def get_stroke_width(self):
        # Turn down stroke width -> Only subtly bevel the wordlist bar.
        value = super(WordlistKey, self).get_stroke_width()
        return min(value, 0.6)

    def get_stroke_gradient(self):
        # Turn down stroke gradient -> Only subtly bevel the wordlist bar.
        value = super(WordlistKey, self).get_stroke_gradient()
        return min(value, 0.3)

    def get_light_direction(self):
        return -0.3 * pi / 180

    def draw_shadow_cached(self, context):
        # no shadow
        pass


class FullSizeKey(WordlistKey):
    def __init__(self, id = "", border_rect = None):
        super(FullSizeKey, self).__init__(id, border_rect)

    def get_rect(self):
        """ Get bounding box in logical coordinates """
        # Disable key_size, let wordlist creation have complete size control.
        return self.get_fullsize_rect()


class FlatKey(FullSizeKey):
    def __init__(self, id="", border_rect=None):
        super(FlatKey, self).__init__(id, border_rect)

    def draw(self, context, lod=LOD.FULL):
        # draw only when pressed, to blend in with the word list bar
        if self.pressed or self.active or self.scanned:
            self.draw_geometry(context, lod)
        self.draw_image(context, lod)
        self.draw_label(context, lod)

    def get_stroke_width(self):
        # Turn down stroke width -> no annoying banding at
        # what should be flat key edges.
        return 0.0


class BarKey(FlatKey):
    def __init__(self, id="", border_rect=None):
        super(BarKey, self).__init__(id, border_rect)

    def can_show_label_popup(self):
        return False


class WordKey(FixedFontMixin, BarKey):
    def __init__(self, id="", border_rect=None):
        super(WordKey, self).__init__(id, border_rect)


class InputlineKey(FixedFontMixin, RectKey, InputlineKeyCommon):

    cursor = 0

    def __init__(self, id="", border_rect = None):
        RectKey.__init__(self, id, border_rect)
        self.word_infos = []
        self._xscroll = 0.0

    def set_content(self, line, word_infos, cursor):
        self.line = line
        self.word_infos = word_infos
        self.cursor = cursor
        self.invalidate_key()

        # determine text direction
        dir = Pango.find_base_dir(line, -1)
        self.ltr = dir != Pango.Direction.RTL

    def draw_label(self, context, lod):
        layout, rect, cursor_rect, layout_pos = self._calc_layout_params()
        cursor_width = cursor_rect.h * 0.075
        cursor_width = max(cursor_width, 1.0)
        label_rgba = self.get_label_color()

        context.save()
        context.rectangle(*rect)
        context.clip()

        # draw text
        context.set_source_rgba(*label_rgba)
        context.move_to(*layout_pos)
        PangoCairo.show_layout(context, layout)

        context.restore()  # don't clip the caret

        # draw caret
        context.move_to(cursor_rect.x, cursor_rect.y)
        context.rel_line_to(0, cursor_rect.h)
        context.set_source_rgba(*label_rgba)
        context.set_line_width(cursor_width)
        context.stroke()

        # reset attributes; layout is reused by all keys due to memory leak
        layout.set_attributes(Pango.AttrList())

    def get_layout(self):
        text, attrs = self._build_layout_contents()
        layout = self.get_pango_layout(text, self.font_size)
        layout.set_attributes(attrs)
        layout.set_auto_dir(True)
        return layout

    def get_canvas_label_rect(self):
        rect = super(InputlineKey, self).get_canvas_label_rect()
        return rect.int()       # else clipping glitches

    def _build_layout_contents(self):
        # Add one char to avoid having to handle RTL corner cases at line end.
        text =  self.line + " "
        attrs = None

        # prepare colors
        color_ignored       = '#00FFFF'
        color_partial_match = '#00AA00'
        color_no_match      = '#00FF00'
        color_error         = '#FF0000'

        # set text colors, highlight unknown words
        #   AttrForeground/pango_attr_foreground_new are still inaccassible
        #   -> use parse_markup instead.
        # https://bugzilla.gnome.org/show_bug.cgi?id=646788
        markup = ""
        wis = self.word_infos
        for i, wi in enumerate(wis):
            cursor_at_word_end = self.cursor == wi.end

            # select colors
            predict_color = None
            spell_color = None
            if 0:  # no more bold, keep it simple
                if wi.ignored:
                    #color = color_ignored
                    pass
                elif not wi.exact_match:
                    if wi.partial_match and cursor_at_word_end:
                        predict_color = color_partial_match
                    else:
                        predict_color = color_no_match

            if wi.spelling_errors:
                spell_color = color_error

            # highlight the word as needed
            word = text[wi.start : wi.end]
            word = GLib.markup_escape_text(word)
            if predict_color or spell_color:
                span = ""
                if predict_color:
                    span += "<b>"
                if spell_color:
                    span += "<span underline_color='" + spell_color + "' " + \
                                 "underline='error'>"
                span += word

                if spell_color:
                    span += "</span>"
                if predict_color:
                    span += "</b>"

                t = span
            else:
                span = word

            # assemble the escaped pieces
            if i == 0:
                # add text up to the first word
                intro = text[:wi.start]
                markup += GLib.markup_escape_text(intro)
            else:
                # add gap between words
                wiprev = wis[i-1]
                gap = text[wiprev.end : wi.start]
                markup += GLib.markup_escape_text(gap)

            # add the word
            markup += span

            if i == len(wis) - 1:
                # add remaining text after the last word
                remainder = text[wi.end:]
                markup += GLib.markup_escape_text(remainder)

        result = Pango.parse_markup(markup, -1, "\0")
        if len(result) == 4:
            ok, attrs, text, error = result

        return text, attrs

    def _calc_layout_params(self):
        layout = self.get_layout()

        # get label rect and aligned drawing origin
        rect = self.get_canvas_label_rect()
        text_size = layout.get_pixel_size()
        xalign, yalign = self.align_label(text_size, (rect.w, rect.h), self.ltr)

        # get cursor position
        cursor_index = self.cursor_to_layout_index(layout, self.cursor, self.ltr)
        strong_pos, weak_pos = layout.get_cursor_pos(cursor_index)
        pos = strong_pos
        cursor_rect = Rect(pos.x, pos.y, pos.width, pos.height).scale(1.0 / Pango.SCALE)

        # scroll to cursor
        self._update_scroll_position(rect, text_size, cursor_rect, xalign)

        xlayout = rect.x + xalign + self._xscroll
        ylayout = rect.y + yalign
        cursor_rect.x += xlayout
        cursor_rect.y += ylayout

        return layout, rect, cursor_rect, (xlayout, ylayout)

    def _update_scroll_position(self, label_rect, text_size,
                                cursor_rect, xalign):
        xscroll = self._xscroll

        # scroll line into view
        gap_begin = xalign + xscroll
        gap_end   = label_rect.w - (xalign + xscroll + text_size[0])
        if gap_begin > 0 or gap_end > 0:
            xscroll = 0

        # scroll cursor into view
        over_begin = -(xalign + cursor_rect.x)
        over_end   =   xalign + cursor_rect.x - label_rect.w
        if over_begin - xscroll > 0.0:
            xscroll = over_begin
        if over_end + xscroll > 0.0:
            xscroll = -over_end

        self._xscroll = xscroll

    @staticmethod
    def cursor_to_layout_index(layout, cursor, ltr = False):
        """ Translate unicode character position to pango byte index. """
        indexes = []
        i = 0
        iter = layout.get_iter()
        while True:
            indexes.append(iter.get_index())
            if not iter.next_char():
                break

        if ltr:
            if len(indexes) == 0:
                cursor_index = 0
            elif cursor < 0:
                cursor_index = 0
            elif cursor >= len(indexes):
                cursor_index = indexes[-1]
            else:
                cursor_index = indexes[cursor]
        else:
            if len(indexes) == 0:
                cursor_index = 0
            elif cursor < 0:
                cursor_index = indexes[-1]
            elif cursor >= len(indexes):
                cursor_index = 0
            else:
                cursor_index = indexes[-(cursor+1)]

        return cursor_index

zf = None
count = 0

class PixBufScaled:
    """
    Workaround for blurry images when window_scaling_factor >1
    """
    _pixbuf = None
    _width = 0
    _height = 0
    _real_width = 0
    _real_height = 0

    @staticmethod
    def from_file_and_size(filename, width, height):
        pixbuf = PixBufScaled()
        pixbuf._load(filename, width, height)
        return pixbuf

    def get_width(self):
        return self._width

    def get_height(self):
        return self._height

    def _load(self, filename, width, height):
        scale = config.window_scaling_factor
        load_width = width * scale
        load_height = height * scale

        self._pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
            filename, load_width, load_height)

        self._real_width = self._pixbuf.get_width()
        self._real_height = self._pixbuf.get_height()
        self._width = self._real_width / scale
        self._height = self._real_height / scale

    def draw(self, context, rect, rgba, image_style):
        """
        Draw the image in the theme's label color.
        Only the alpha channel of the image is used.
        """
        context.save()

        context.translate(rect.x, rect.y)
        scale = config.window_scaling_factor
        if scale and scale != 1.0:
            context.scale(1.0 / scale, 1.0 / scale)

        # colored?
        if image_style == ImageStyle.MULTI_COLOR:
            Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
            context.paint()

        # grayscale?
        elif image_style == ImageStyle.DESATURATED:
            # gdk_pixbuf_saturate_and_pixelate ()
            if 0:
                # must have non-black background
                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                pattern = context.get_source()
                k = 1 / 256.0
                context.set_source_rgb(k, k, k)
                context.mask(pattern)

                # cairo.OPERATOR_HSL_LUMINOSITY doesn't exist in python bindings
                # CAIRO_OPERATOR_HSL_LUMINOSITY = 28
                context.set_operator(28)
                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                context.paint()
            elif 0:
                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                context.paint()

                import colorsys
                rgb = colorsys.hls_to_rgb(0, 0.5, 0.2)
                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                pattern = context.get_source()

                context.set_source_rgb(*rgb)
                context.set_operator(26)
                context.paint()
            else:

                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                pattern = context.get_source()

                context.push_group_with_content(cairo.CONTENT_COLOR_ALPHA)

                Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
                context.paint()

                import colorsys
                saturation = 0.3
                rgb = colorsys.hls_to_rgb(0, 0.5, saturation)

                context.set_source_rgb(*rgb)
                context.set_operator(26)
                context.paint()

                context.pop_group_to_source()
                context.mask(pattern)

        # single color
        else:
            Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
            pattern = context.get_source()
            context.set_source_rgba(*rgba)
            context.mask(pattern)

        context.restore()





#!/usr/bin/python3

# Copyright © 2010, 2012-2014 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
import unittest
from Onboard.pypredict import *


class _TestPatterns(unittest.TestCase):

    def __init__(self, test, text, result):
        unittest.TestCase.__init__(self, test)
        self.text = text
        self.result = result

    def test_sentence_pattern(self):
        result = SENTENCE_PATTERN.findall(self.text)
        self.assertEqual(result, self.result,
                         "test '%s': '%s' != '%s'" %
                         (self.text, repr(result), repr(self.result)))


class _TestTokenization(unittest.TestCase):

    def __init__(self, test, text, result):
        unittest.TestCase.__init__(self, test)
        self.training_text = text
        self.result = result

    def test_tokenize_text(self):
        tokens, spans = tokenize_text(self.training_text)
        self.assertEqual(tokens, self.result,
                         "test '%s': '%s' != '%s'" %
                         (self.training_text, repr(tokens), repr(self.result)))

    def test_tokenize_context(self):
        tokens, spans = tokenize_context(self.training_text)
        self.assertEqual(tokens, self.result,
                         "test '%s': '%s' != '%s'" %
                         (self.training_text, repr(tokens), repr(self.result)))

    def test_split_sentences(self):
        sentences, spans = split_sentences(self.training_text)
        self.assertEqual(sentences, self.result,
                         "test '%s': '%s' != '%s'" %
                         (self.training_text, repr(sentences), repr(self.result)))


class _TestMultiOrder(unittest.TestCase):
    def __init__(self, test, order):
        unittest.TestCase.__init__(self, test)
        self.order = order

    def setUp(self):
        # text snippets from MOBY DICK By Herman Melville from Project Gutenberg
        self.training_text = """
            No, when I go to sea, I go as a simple sailor, right before the mast,
            plumb down into the forecastle, aloft there to the royal mast-head.
            True, they rather order me about some, and make me jump from spar to
            spar, like a grasshopper in a May meadow. And at first, this sort
            of thing is unpleasant enough. And more than all,
            if just previous to putting your hand into the tar-pot, you have been
            lording it as a country schoolmaster, making the tallest boys stand
            in awe of you. The transition is a keen one, I assure you, from a
            schoolmaster to a sailor, and requires a strong decoction of Seneca and
            the Stoics to enable you to grin and bear it. But even this wears off in
            time.
            """
        self.testing_text = """
            I now took the measure of the bench, and found that it was a foot too
            short; but that could be mended with a chair. I then placed the
            first bench lengthwise along the only clear space against the wall,
            leaving a little interval between, for my back to settle down in. But I
            soon found that there came such a draught of cold air over me from under
            the sill of the window, that this plan would never do at all, especially
            as another current from the rickety door met the one from the window,
            and both together formed a series of small whirlwinds in the immediate
            vicinity of the spot where I had thought to spend the night.
            """
        #self.training_text = u"Mary has a little lamb. Mary has a little lamb."
        #self.training_text = self.testing_text = u"a <s>"
        #self.training_text = self.testing_text = u"a b <s> c"
        #self.training_text = self.testing_text = u"a b c"
        self.training_tokens, _spans = tokenize_text(self.training_text)
        self.testing_tokens, _spans = tokenize_text(self.testing_text)

    def probability_sum(self, model):
        def print(s=""):
            sys.stderr.write(s + '\n')
        # test sum of probabilities for multiple predictions
        num_tests = 0
        num_bad = 0
        num_with_zero = 0

       # import traceback
       # traceback.print_stack()
        for i,t in enumerate(self.testing_tokens):
            context = self.testing_tokens[:i] + [""]
            choices = model.predictp(context,
                                     options = model.NORMALIZE |
                                               model.INCLUDE_CONTROL_WORDS)
            psum = sum(x[1] for x in choices)

            num_tests += 1
            eps = 1e-6

            if abs(1.0 - psum) > eps:
                num_bad += 1
                if num_bad == 1:
                    print()
                print("order %d, pos %d: probabilities don't sum to 1.0; psum=%10f, #results=%6d, context='%s'" % \
                      (self.order, num_tests, psum, len(choices), repr(context[-4:])))

            zerocount = sum(1 for word,p in choices if p == 0)
            if zerocount:
                num_with_zero += 1
                print("order %d, pos %d: %d words with zero probability; psum=%10f, #results=%6d, context='%s'" % \
                      (self.order, num_tests, zerocount, psum, len(choices), repr(context[-4:])))

        self.assertEqual(num_tests, num_tests-num_bad,
                      "order %d, probabilities don't sum to 1.0 for %d of %d predictions" % \
                      (self.order, num_bad, num_tests))

        self.assertEqual(num_tests, num_tests-num_with_zero,
                      "order %d, zero probabilities in %d of %d predictions" % \
                      (self.order, num_with_zero, num_tests))

class _TestMultiOrderMisc(_TestMultiOrder):

    def test_psum_unigram_model(self):
        model = UnigramModel(self.order)
        model.learn_tokens(self.training_tokens)
        self.probability_sum(model)

    def test_psum_dynamic_model_witten_bell(self):
        model = DynamicModel(self.order)
        model.smoothing = "witten-bell"
        model.learn_tokens(self.training_tokens)
        self.probability_sum(model)

    def test_psum_dynamic_model_absolute_discounting(self):
        model = DynamicModel(self.order)
        model.smoothing = "abs-disc"
        model.learn_tokens(self.training_tokens)
        self.probability_sum(model)

    def test_psum_dynamic_model_kneser_ney(self):
        model = DynamicModelKN(self.order)
        model.smoothing = "kneser-ney"
        model.learn_tokens(self.training_tokens)
        self.probability_sum(model)

    def test_psum_cached_dynamic_model(self):
        model = CachedDynamicModel(self.order)
        model.smoothing = "abs-disc"
        model.learn_tokens(self.training_tokens)
        self.probability_sum(model)

    def test_psum_overlay_model(self): # this sums to 1.0 only for identical models
        model = DynamicModel(self.order)
        model.learn_tokens(self.training_tokens)
        self.probability_sum(overlay([model, model]))

    def test_psum_linint_model(self):
        model = DynamicModel(self.order)
        model.learn_tokens(self.training_tokens)
        self.probability_sum(linint([model, model]))

    def test_psum_loglinint_model(self):
        model = DynamicModel(self.order)
        model.learn_tokens(self.training_tokens)
        self.probability_sum(loglinint([model, model]))

    def test_prune_witten_bell(self):
        model = DynamicModel(self.order)
        model.learn_tokens(self.training_tokens)
        for prune_count in range(5):
            m = model.prune([prune_count])
            m.smoothing = "witten-bell"
            self.probability_sum(m)

    def test_prune_absolute_discounting(self):
        model = DynamicModel(self.order)
        model.learn_tokens(self.training_tokens)
        for prune_count in range(5):
            m = model.prune([prune_count])
            m.smoothing = "abs-disc"
            self.probability_sum(m)

    def test_prune_kneser_ney(self):
        model = DynamicModelKN(self.order)
        model.learn_tokens(self.training_tokens)
        for prune_count in range(5):
            m = model.prune([prune_count])
            m.smoothing = "kneser-ney"
            self.probability_sum(m)


class _TestMultiOrderLoadingRobustness(_TestMultiOrder):

    def setUp(self):
        self._tmp_dir = tempfile.TemporaryDirectory(prefix="test_onboard_")
        self._dir = self._tmp_dir.name

        # Enough text for at least 3 n-grams of max-order 5.
        text = "word1 word2 word3 word4 word5 word6 word7"
        tokens, _spans = tokenize_text(text)

        # prepare contents of error-free model
        order = self.order
        fn = os.path.join(self._dir, "order{}.lm".format(order))
        if order == 1:
            model = UnigramModel()
        else:
            model = DynamicModel(order)
        model.learn_tokens(tokens)
        model.save(fn)

        with open(fn, encoding="UTF-8") as f:
            lines = f.readlines()

        self._model = model
        self._model_contents = [fn, lines]

    def test_load_models_with_bad_control_word_counts(self):
        """
        Control words with counts <= 0 must be repaired on load.
        """
        model = self._model
        order = self.order
        counts_before = model.get_counts()

        for field_change in ["-1000, -1", "0"]:
            fn, lines = self._model_contents
            index = lines.index("\\1-grams:\n")
            nlines = []
            for i, line in enumerate(lines):
                if i > index and i < index + 1 + model.NUM_CONTROL_WORDS:
                    fields = line.split()
                    fields[0] = field_change
                    line = " ".join(fields) + "\n"
                nlines.append(line)

            self._write_contents(fn, nlines)

            # throws IOError
            model.load(fn)

            # Counts of control words must have been corrected to 1 after loading.
            contents = [x for x in model.iter_ngrams()]
            self.assertEqual(contents[:4],
                            [(('<unk>',), 1, 0),
                            (('<s>',), 1, 0),
                            (('</s>',), 1, 0),
                            (('<num>',), 1, 0)])

    def test_load_models_with_bad_ngram_counts(self):
        """
        N-grams with counts <= 0 must be ignored on load.
        """
        model = self._model
        order = self.order
        counts_before = model.get_counts()

        for field_change in ["-1000", "-1", "0"]:
            fn, lines = self._model_contents
            nlines = []
            count = None
            lineno = 0
            for line in lines:
                lineno += 1
                if "1-grams:" in line:
                    count = -model.NUM_CONTROL_WORDS   # skip control words
                elif "-grams:" in line:
                    count = 0
                if count == 2:
                    count = None
                    fields = line.split()
                    fields[0] = field_change
                    line = " ".join(fields) + "\n"
                if not count is None:
                    count += 1
                nlines.append(line)

            self._write_contents(fn, nlines)

            # throws IOError
            model.load(fn)

            msg = "order {}, field_change '{}'".format(order, field_change)

            # Counts of control words must have been corrected to 1 after loading.
            contents = [x for x in model.iter_ngrams()]
            self.assertEqual(contents[:4],
                            [(('<unk>',), 1, 0),
                            (('<s>',), 1, 0),
                            (('</s>',), 1, 0),
                            (('<num>',), 1, 0)],
                             msg)

            # Verify count of the valid original model.
            self.assertEqual(counts_before[0],
                             [11, 6, 5, 4, 3][:order],
                             msg)

            # All other n-grams with bad counts must have been dropped.
            self.assertEqual(model.get_counts()[0],
                             [10, 5, 4, 3, 2][:order],
                             msg)

            # No unexpected counts must sneek in, all 1 here.
            for it in model.iter_ngrams():
                ngram = it[0]
                count = it[1]
                self.assertEqual(count, 1, "order {}, field_change '{}': "
                                           "count mismatch in {}" \
                                           .format(order, field_change,
                                                   [ngram, count]))

    def _write_contents(self, fn, lines):
        with open(fn, mode="w", encoding="UTF-8") as f:
            for l in lines:
                f.write(l)


class _TestMultiOrderRemove(_TestMultiOrder):
    tests = \
    [
        # test #0: edge case empty model
        [
            [""],                      # training texts
            [
                [ [["a"]], [""], ],    # removals, testing texts
            ]
        ],
        # test #1-2: edge case single n-gram
        [
            ["a"],
            [
                [ [["a"]], [""], ],
                [ [["f"]], ["a"], ],
            ]
        ],
        # test #3-4: edge case single word type
        [
            ["a", "a a", "a a a"],
            [
                [ [["a"]], [""], ],
                [ [["f"]], ["a", "a a", "a a a"], ],
            ]
        ],
        # test #5-10: remove any single token
        [
            ["a b c d e"],
            [
                [ [["a"]], ["b c d e"], ],
                [ [["b"]], ["a", "c d e"], ],
                [ [["c"]], ["a b", "d e"], ],
                [ [["d"]], ["a b c", "e"], ],
                [ [["e"]], ["a b c d"], ],
                [ [["f"]], ["a b c d e"], ],
            ]
        ],
        # test #11-16: remove any single word of more complex text
        [
            ["a b c d e a b c d e e d c b a", "a b c d e"],
            [
                [ [["a"]], ["b c d e", "b c d e e d c b", "b c d e"], ],
                [ [["b"]], ["a", "c d e a", "c d e e d c", "a", "a", "c d e"], ],
                [ [["c"]], ["a b", "d e a b", "d e e d", "b a", "a b", "d e"], ],
                [ [["d"]], ["a b c", "e a b c", "e e", "c b a", "a b c", "e"], ],
                [ [["e"]], ["a b c d", "a b c d", "d c b a", "a b c d"], ],
                [ [["f"]], ["a b c d e a b c d e e d c b a", "a b c d e"], ],
                [ [["a"], ["b"]], ["c d e", "c d e e d c", "c d e"], ],
                [ [["a"], ["b"], ["c"], ["d"], ["e"]], [""], ],
            ]
        ],
        # remove word occuring after history of size 1, simple
        [
            ["a b c d e"],
            [
                [ [["a", "b"]], ["a", "b c d e"], ],
                [ [["b", "c"]], ["a b", "c d e"], ],
                [ [["c", "d"]], ["a b c", "d e"], ],
                [ [["d", "e"]], ["a b c d", "e"], ],
                [ [["a", "c"]], ["a b c d e"], ],
            ],
        ],
        # remove word occuring after history of size 1, multiple occurences
        [
            ["a", "a b", "a b c", "a b c d", "a b c d e"],
            [
                [ [["a", "b"]],
                  ["a", "a", "b", "a", "b c", "a", "b c d", "a", "b c d e"], ],
            ]
        ],
        # remove word occuring after history of size 1, away from text begin
        [
            ["b a", "b a b", "b a b c", "b a b c d", "b a b c d e"],
            [
                [ [["a", "b"]],
                  ["b a", "b a", "b", "b a", "b c", "b a", "b c d", "b a", "b c d e"], ],
            ]
        ],

        # remove word occuring after history of size 2, simple
        [
            ["a b c d e"],
            [
#                [ [["a", "b", "c"]], ["a b", "c d e"], ],  # not supported yet
            ],
        ],
    ]

    def test_remove_context_control_words(self):
        """ Control words must not be removed """
        control_words = ["<unk>", "<s>", "</s>", "<num>"]
        model = DynamicModel()
        for w in control_words:
            model.count_ngram([w], 1) # Use count_ngram because learn_tokens
                                      # removes some control words.

        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
                         [(('<unk>',), 2, 0),
                          (('<s>',), 2, 0),
                          (('</s>',), 2, 0),
                          (('<num>',), 2, 0)])

        # remove control words, counts must not go below 1
        for w in control_words:
            model.remove_context([w])

        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
                         [(('<unk>',), 1, 0),
                          (('<s>',), 1, 0),
                          (('</s>',), 1, 0),
                          (('<num>',), 1, 0)])

        # again, counts must still be at least 1
        for w in control_words:
            model.remove_context([w])

        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
                         [(('<unk>',), 1, 0),
                          (('<s>',), 1, 0),
                          (('</s>',), 1, 0),
                          (('<num>',), 1, 0)])

    def test_remove_context_save(self):
        """ Removed n-grams must not appear in saved model files """
        tmp_dir = tempfile.TemporaryDirectory(prefix="test_onboard_")
        fn = os.path.join(tmp_dir.name, "remove_context.lm")

        tests = \
        [
            [
                ["a b c <s> d e f"],        # training texts
                [
                    # removals, saved n-gram counts per level (up to order 5)
                    [ [["a"], ["c"], ["f"]], [7, 2, 1, 0, 0] ],
                ]
            ],
            [
                ["a b c d e a b c d e e d c b a", "a b c d e"],
                [
                    [ [["a"]], [8, 7, 6, 5, 4] ],
                    [ [["b"]], [8, 6, 5, 4, 2] ],
                    [ [["c"]], [8, 6, 4, 2, 0] ],
                    [ [["d"]], [8, 6, 3, 1, 0] ],
                    [ [["e"]], [8, 6, 4, 2, 0] ],
                    [ [["f"]], [9, 10, 10, 10, 10] ],
                    [ [["a"], ["b"], ["c"], ["d"], ["e"]], [4, 0, 0, 0, 0] ],
                ]
            ],
        ]

        itest = 0
        for training_texts, test_set in tests:
            for remove_contexts, ngram_counts in test_set:
                model = DynamicModel(self.order)
                for text in training_texts:
                    tokens, spans = tokenize_text(text)
                    model.learn_tokens(tokens)
                for context in remove_contexts:
                    model.remove_context(context)

                model.save(fn)

                # check if it loads, throws exception if it detects anything funny
                model.load(fn)

                # check n-gram counts directly in the saved file
                with open(fn) as f:
                    contents = f.read()
                #print(contents, file=sys.stderr)

                for level in range(self.order):
                    pattern = "ngram {}={}".format(level+1, ngram_counts[level])
                    matches = re.findall(pattern, contents, re.MULTILINE)
                    self.assertEqual(1, len(matches),
                        "order {}, test {}: saved model doesn't contain '{}'" \
                        .format(self.order, itest, pattern))

                itest += 1

    def test_remove_context(self):
        """
        remove_context() must fully perform the removal and leave models in
        a state that can be reached by learning.
        """
        itest = 0
        for training_texts, test_set in self.tests:
            for remove_contexts, testing_texts in test_set:
                self._test_remove_context(
                    training_texts, testing_texts, remove_contexts,
                    self.order, "test #{}".format(itest))
                itest += 1

    def test_get_remove_context_changes(self):
        """
        get_remove_context_changes() must return valid and complete changes.
        """
        itest = 0
        for training_texts, test_set in self.tests:
            for remove_contexts, testing_texts in test_set:
                self._test_get_remove_context_changes(
                    training_texts, testing_texts, remove_contexts,
                    self.order, "test #{}".format(itest))
                itest += 1

    def test_remove_context_witten_bell(self):
        """
        Witten-bell predictions must sum to zero after removal.
        """
        for training_texts, test_set in self.tests:
            for remove_contexts, testing_texts in test_set:

                model = DynamicModel(self.order)
                model.smoothing = "witten-bell"

                for text in training_texts:
                    tokens, _spans = tokenize_text(text)
                    model.learn_tokens(tokens)

                # remove
                for context in remove_contexts:
                    model.remove_context(context)
                self.probability_sum(model)

                # re-add
                for context in remove_contexts:
                    model.learn_tokens(context)
                self.probability_sum(model)

    def test_remove_context_absolute_discounting(self):
        """
        Absolute-discounting predictions must sum to zero after removal.
        """
        for training_texts, test_set in self.tests:
            for remove_contexts, testing_texts in test_set:

                model = DynamicModel(self.order)
                model.smoothing = "abs-disc"
                for text in training_texts:
                    tokens, _spans = tokenize_text(text)
                    model.learn_tokens(tokens)

                # remove
                for context in remove_contexts:
                    model.remove_context(context)
                self.probability_sum(model)

                # re-add
                for context in remove_contexts:
                    model.learn_tokens(context)
                self.probability_sum(model)

    def test_remove_context_kneser_ney(self):
        """
        Kneser-ney predictions must sum to zero after removal.
        """
        for training_texts, test_set in self.tests:
            for remove_contexts, testing_texts in test_set:

                model = DynamicModelKN(self.order)
                model.smoothing = "kneser-ney"
                for text in training_texts:
                    tokens, _spans = tokenize_text(text)
                    model.learn_tokens(tokens)

                for context in remove_contexts:
                    model.remove_context(context)

                # remove
                for context in remove_contexts:
                    model.remove_context(context)
                self.probability_sum(model)

                # re-add
                for context in remove_contexts:
                    model.learn_tokens(context)
                self.probability_sum(model)

    def _test_remove_context(self, training_texts, testing_texts,
                                   remove_contexts, order, description):
        model0 = DynamicModel(order)
        for text in training_texts:
            tokens, _spans = tokenize_text(text)
            model0.learn_tokens(tokens)

        model1 = DynamicModel(order)
        for text in testing_texts:
            tokens, _spans = tokenize_text(text)
            model1.learn_tokens(tokens)

        for context in remove_contexts:
            model0.remove_context(context)
        ngrams0 = sorted(sorted(model0.iter_ngrams()), key=lambda x: -len(x[0]))
        ngrams1 = sorted(sorted(model1.iter_ngrams()), key=lambda x: -len(x[0]))

        # compare testing model with results of simulated removal
        for i, ng in enumerate(ngrams0):
            ng0 = [ngrams0[i][0], ngrams0[i][1]]
            if i < len(ngrams1):
                ng1 = [ngrams1[i][0], ngrams1[i][1]]
            else:
                ng1 = None

            self.assertEqual(ng0, ng1,
                             "{}, order {}: n-gram mismatch:  {} != {}" \
                             .format(description, order, ng0, ng1))

    def _test_get_remove_context_changes(self, training_texts, testing_texts,
                                         remove_contexts, order, description):
        model0 = DynamicModel(order)
        for text in training_texts:
            tokens, _spans = tokenize_text(text)
            model0.learn_tokens(tokens)

        model1 = DynamicModel(order)
        for text in testing_texts:
            tokens, _spans = tokenize_text(text)
            model1.learn_tokens(tokens)

        ngrams0 = list(model0.iter_ngrams())
        ngrams1 = list(model1.iter_ngrams())

        # simulate removal by applying the required changes to ngrams0
        changes = {}
        for context in remove_contexts:
            cs = model0.get_remove_context_changes(context)
            for ngram, count in cs.items():
                changes[ngram] = changes.get(ngram, 0) + count

        ngrams0_mod = []
        for i, it in enumerate(ngrams0):
            ngram = it[0]
            count = it[1]
            diff = changes.get(ngram, 0)
            count += diff

            # Test results in negative counts for more than one
            # removal. This is not a fault of get_remove_context_changes()
            # and doesn't happen with remove_context() either.
            if len(remove_contexts) == 1 and count != 0 or \
               len(remove_contexts) >= 2 and count > 0:
                ngrams0_mod.append([ngram, count])

        ngrams0 = sorted(sorted(ngrams0), key=lambda x: -len(x[0]))
        ngrams0_mod = sorted(sorted(ngrams0_mod), key=lambda x: -len(x[0]))
        ngrams1 = sorted(sorted(ngrams1), key=lambda x: -len(x[0]))

        if 0:
            if order == 2:
                for ng in ngrams0:
                    print("0:", ng[0], ng[1], file=sys.stderr)
                print(file=sys.stderr)
                for ng in ngrams0_mod:
                    print("m:", ng[0], ng[1], file=sys.stderr)
                print(file=sys.stderr)
                for ng in ngrams1:
                    print("1:", ng[0], ng[1], file=sys.stderr)
                print(file=sys.stderr)

        # compare testing model with results of simulated removal
        for i, ng in enumerate(ngrams0_mod):
            ng0 = [ngrams0_mod[i][0], ngrams0_mod[i][1]]
            if i < len(ngrams1):
                ng1 = [ngrams1[i][0], ngrams1[i][1]]
            else:
                ng1 = None

            self.assertEqual(ng0, ng1,
                             "{}, order {}: n-gram mismatch:  {} != {}" \
                             .format(description, order, ng0, ng1))


class _TestModel(unittest.TestCase):

    def setUp(self):
        self._tmp_dir = tempfile.TemporaryDirectory(prefix="test_onboard_")
        self._dir = self._tmp_dir.name

    def test_case_insensitive(self):
        model = DynamicModel()
        model.learn_tokens(['ABCDE'], 1)

        choices = model.predict(['a'])
        self.assertEqual(choices, [])

        choices = model.predict(['abcde'], options = model.CASE_INSENSITIVE)
        self.assertEqual(choices, ['ABCDE'])

    def test_accent_insensitive(self):
        model = DynamicModel()
        model.learn_tokens(['ÉéÈèñ'], 1)
        model.learn_tokens(['früh', 'fruchtig'], 1)

        choices = model.predict(['EeEen'])
        self.assertEqual(choices, [])

        choices = model.predict(['EeEen'], options = model.ACCENT_INSENSITIVE)
        self.assertEqual(choices, ['ÉéÈèñ'])

    def test_accent_insensitive_smart(self):
        model = DynamicModel()
        model.learn_tokens(['früh', 'fruchtig'], 1)

        choices = model.predict(['fru'])
        self.assertEqual(choices, ['fruchtig'])

        choices = model.predict(['fru'], options = model.ACCENT_INSENSITIVE_SMART)
        self.assertEqual(choices, ['früh', 'fruchtig'])

        choices = model.predict(['frü'], options = model.ACCENT_INSENSITIVE_SMART)
        self.assertEqual(choices, ['früh'])

    def test_ignore_capitalized(self):
        model = DynamicModel()
        model.learn_tokens(['ABCDE'], 1)
        model.learn_tokens(['abcde'], 1)

        choices = model.predict([''])
        self.assertEqual(choices, ['ABCDE', 'abcde'])

        choices = model.predict([''], options = model.IGNORE_CAPITALIZED)
        self.assertEqual(choices, ['abcde'])

    def test_ignore_non_capitalized(self):
        model = DynamicModel()
        model.learn_tokens(['ABCDE'], 1)
        model.learn_tokens(['abcde'], 1)

        choices = model.predict([''])
        self.assertEqual(choices, ['ABCDE', 'abcde'])

        choices = model.predict([''], options = model.IGNORE_NON_CAPITALIZED)
        self.assertEqual(choices, ['ABCDE'])

    def test_save_load_unigram_model(self):
        fn = os.path.join(self._dir, "unigram.lm")

        model = UnigramModel()
        tokens = tokenize_text("ccc bbb uu fff ccc ee")[0]
        model.learn_tokens(tokens)
        model.save(fn)

        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
        [(('<unk>',), 1),
         (('<s>',), 1),
         (('</s>',), 1),
         (('<num>',), 1),
         (('ccc',), 2),
         (('bbb',), 1),
         (('uu',), 1),
         (('fff',), 1),
         (('ee',), 1)]
        )

        # Loading should sort unigrams except the initial control words.
        # Reasons: - Obfuscation of the learned text on second save.
        #          - Making Dictionary::sorted redundant to save memory
        #            and improve performance by working around its insert
        #            inefficiency (becomes crippling with very large
        #            vocabularies, i.e. millions of words)
        model = UnigramModel()
        model.load(fn)
        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
        [(('<unk>',), 1),
         (('<s>',), 1),
         (('</s>',), 1),
         (('<num>',), 1),
         (('bbb',), 1),
         (('ccc',), 2),
         (('ee',), 1),
         (('fff',), 1),
         (('uu',), 1)]
        )

    def test_save_load_trigram_model(self):
        fn = os.path.join(self._dir, "unigram.lm")

        model = DynamicModel()
        tokens = tokenize_text("ccc bbb uu fff ccc ee")[0]
        model.learn_tokens(tokens)
        model.save(fn)

        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
            [(('<unk>',), 1, 0),
             (('<s>',), 1, 0),
             (('</s>',), 1, 0),
             (('<num>',), 1, 0),
             (('ccc',), 2, 2),
             (('ccc', 'bbb'), 1, 1),
             (('ccc', 'bbb', 'uu'), 1, 0),
             (('ccc', 'ee'), 1, 0),
             (('bbb',), 1, 1),
             (('bbb', 'uu'), 1, 1),
             (('bbb', 'uu', 'fff'), 1, 0),
             (('uu',), 1, 1),
             (('uu', 'fff'), 1, 1),
             (('uu', 'fff', 'ccc'), 1, 0),
             (('fff',), 1, 1),
             (('fff', 'ccc'), 1, 1),
             (('fff', 'ccc', 'ee'), 1, 0),
             (('ee',), 1, 0)]
        )

        # Loading should sort unigrams except the initial control words.
        model = DynamicModel()
        model.load(fn)
        contents = [x for x in model.iter_ngrams()]
        self.assertEqual(contents,
            [(('<unk>',), 1, 0),
             (('<s>',), 1, 0),
             (('</s>',), 1, 0),
             (('<num>',), 1, 0),
             (('bbb',), 1, 1),
             (('bbb', 'uu'), 1, 1),
             (('bbb', 'uu', 'fff'), 1, 0),
             (('ccc',), 2, 2),
             (('ccc', 'bbb'), 1, 1),
             (('ccc', 'bbb', 'uu'), 1, 0),
             (('ccc', 'ee'), 1, 0),
             (('ee',), 1, 0),
             (('fff',), 1, 1),
             (('fff', 'ccc'), 1, 1),
             (('fff', 'ccc', 'ee'), 1, 0),
             (('uu',), 1, 1),
             (('uu', 'fff'), 1, 1),
             (('uu', 'fff', 'ccc'), 1, 0)]
        )

    def test_read_order(self):
        """ Test reading the order of a language model """
        fn = os.path.join(self._dir, "model.lm")

        self.assertEqual(read_order(fn), None) # file not found

        model = UnigramModel()
        tokens = tokenize_text("ccc bbb uu fff ccc ee")[0]
        model.learn_tokens(tokens)
        model.save(fn)
        self.assertEqual(read_order(fn), 1)

        model = DynamicModel()
        tokens = tokenize_text("ccc bbb uu fff ccc ee")[0]
        model.learn_tokens(tokens)
        model.save(fn)
        self.assertEqual(read_order(fn), 3)


def suite():

    # input-text, text-tokens, context-tokens, sentences
    tests = [
         ["", [], [], []],
         ["abc", ["abc"], ["abc"], ["abc"]],
         ["We saw wha", ['We', 'saw', 'wha'], ['We', 'saw', 'wha'],
             ['We saw wha']],
         ["We saw whales", ['We', 'saw', 'whales'],
             ['We', 'saw', 'whales'],
             ['We saw whales']],
         ["We saw whales ", ['We', 'saw', 'whales'],
             ['We', 'saw', 'whales', ''],
             ['We saw whales']],
         ["We  saw     whales", ['We', 'saw', 'whales'],
             ['We', 'saw', 'whales'],
             ['We  saw     whales']],
         ["Hello there! We saw whales ",
             ['Hello', 'there', '<s>', 'We', 'saw', 'whales'],
             ['Hello', 'there', '<s>', 'We', 'saw', 'whales', ''],
             ['Hello there!', 'We saw whales']],
         ["Hello there! We saw 5 whales ",
             ['Hello', 'there', '<s>', 'We', 'saw', '<num>', 'whales'],
             ['Hello', 'there', '<s>', 'We', 'saw', '<num>', 'whales', ''],
             ['Hello there!', 'We saw 5 whales']],
         ["Hello there! We #?/=$ saw 5 whales ",
             ['Hello', 'there', '<s>', 'We', 'saw', '<num>', 'whales'],
             ['Hello', 'there', '<s>', 'We', 'saw', '<num>', 'whales', ''],
             ['Hello there!', 'We #?/=$ saw 5 whales']],
         [".", [], [''], ['.']],
         [". ", ['<s>'], ['<s>', ''], ['.', '']],
         [". sentence.", ['<s>', 'sentence'], ['<s>', 'sentence', ''],
             ['.', 'sentence.']],
         ["sentence.", ['sentence'], ['sentence', ''], ['sentence.']],
         ["sentence. ", ['sentence', '<s>'], ['sentence', '<s>', ''],
             ['sentence.', '']],
         ["sentence. sentence.", ['sentence', '<s>', 'sentence'],
             ['sentence', '<s>', 'sentence', ''],
             ['sentence.', 'sentence.']],
         ["sentence. sentence. ", ['sentence', '<s>', 'sentence', '<s>'],
             ['sentence', '<s>', 'sentence', '<s>', ''],
             ['sentence.', 'sentence.', '']],
         ["sentence.\n sentence. ", ['sentence', '<s>', 'sentence', '<s>'],
             ['sentence', '<s>', 'sentence', '<s>', ''],
             ['sentence.', 'sentence.', '']],
         ["sentence. \nsentence.", ['sentence', '<s>', 'sentence'],
             ['sentence', '<s>', 'sentence', ''],
             ['sentence.', 'sentence.']],
         ["sentence. \n", ['sentence', '<s>'], ['sentence', '<s>', ''],
             ['sentence.', '']],
         ['sentence "quote." sentence.',
             ['sentence', 'quote', '<s>', 'sentence'],
             ['sentence', 'quote', '<s>', 'sentence', ''],
             ['sentence "quote."', 'sentence.']],
         ["sentence <s>", ['sentence'], ['sentence', ''], ['sentence']],
         [""""double quotes" 'single quotes'""",
             ['double', 'quotes', 'single', "quotes'"],
             ['double', 'quotes', 'single', "quotes'"],
             ['"double quotes" \'single quotes\'']],
         ["(parens) [brackets] {braces}",
             ['parens', 'brackets', 'braces'],
             ['parens', 'brackets', 'braces', ''],
             ['(parens) [brackets] {braces}']],
         ["\nnewline ", ['newline'], ['newline', ''], ['newline']],
         ["double\n\nnewline ", ['double', '<s>', 'newline'],
             ['double', '<s>', 'newline', ''], ['double', 'newline']],
         ["double_newline\n\n", ['double_newline', '<s>'],
             ['double_newline', '<s>', ''], ['double_newline', '']],
         ["double_newline \n \n \n", ['double_newline', '<s>'],
             ['double_newline', '<s>', ''], ['double_newline', '']],
         ["dash-dash", ["dash-dash"], ["dash-dash"], ["dash-dash"]],
         ["dash-", ['dash'], ['dash-'], ['dash-']],
         ["single quote's", ['single', "quote's"], ['single', "quote's"],
             ["single quote's"]],
         ["single quote'", ['single', "quote'"], ['single', "quote'"],
             ["single quote'"]],
         ["under_score's", ["under_score's"], ["under_score's"],
             ["under_score's"]],
         ["Greek Γ´", ['Greek', 'Γ´'], ['Greek', 'Γ´'], ["Greek Γ´"]], # U+00b4
         ["Greek Γ΄", ['Greek', 'Γ΄'], ['Greek', 'Γ΄'], ["Greek Γ΄"]], # U+0384

         # command line handling
         ["-option --option", ['-option', '--option'], ['-option', '--option'],
             ['-option --option']],
         ["cmd -", ['cmd'], ['cmd', '-'], ['cmd -']],
         ["cmd - ", ['cmd'], ['cmd', '-', ''], ['cmd -']],
         ["cmd --", ['cmd'], ['cmd', '--'], ['cmd --']],
         ["cmd -- ", ['cmd'], ['cmd', '--', ''], ['cmd --']],
         ["cmd ---", ['cmd', '<unk>'], ['cmd', '<unk>'], ['cmd ---']],
         ["|", ['|'], ['|'], ['|']],
         ["find | grep", ['find', '|', 'grep'], ['find', '|', 'grep'],
             ['find | grep']],
         ["cat /", ['cat'], ['cat', ''], ['cat /']],

         # passing through control words
         ["<unk> <s> </s> <num>", ['<unk>', '<s>', '</s>', '<num>'],
             ['<unk>', '<s>', '</s>', '<num>', ''],
             ['<unk>', '</s> <num>']],

         # <unk>
         ["repeats: a aa aaa aaaa aaaaa",
             ['repeats', '<s>', 'a', 'aa', 'aaa', '<unk>', '<unk>'],
             ['repeats', '<s>', 'a', 'aa', 'aaa', '<unk>', '<unk>'],
             ['repeats:', 'a aa aaa aaaa aaaaa']],

         # <num>
         ["1", ["<num>"], ["<num>"], ["1"]],
         ["123", ['<num>'], ['<num>'], ["123"]],
         ["-3", ["<num>"], ["<num>"], ["-3"]],
         ["+4", ["<num>"], ["<num>"], ["+4"]],
         ["123.456", ["<num>"], ["<num>"], ["123.456"]],
         ["123,456", ["<num>"], ["<num>"], ["123,456"]],
         ["100,000.00", ["<num>"], ["<num>"], ["100,000.00"]],
         ["100.000,00", ["<num>"], ["<num>"], ["100.000,00"]],
         [".5", ["<num>"], ["<num>"], [".5"]],

         # begin of text markers
         ["<bot:txt> word", ['<bot:txt>', 'word'], ['<bot:txt>', 'word'],
             ['<bot:txt> word']],
         ["<bot:term> word", ['<bot:term>', 'word'], ['<bot:term>', 'word'],
             ['<bot:term> word']],
         ["<bot:url> word", ['<bot:url>', 'word'], ['<bot:url>', 'word'],
             ['<bot:url> word']],
         ["<bot:txt> sentence. sentence. ",
             ['<bot:txt>', 'sentence', '<s>', 'sentence', '<s>'],
             ['<bot:txt>', 'sentence', '<s>', 'sentence', '<s>', ''],
             ['<bot:txt> sentence.', 'sentence.', '']],

         # URLs
         ["www",  ['www'], ['www'],     ['www']],
         ["www ", ['www'], ['www', ''], ['www']],
         ["www.", ['www'], ['www', ''], ['www.']],
         ["www,", ['www'], ['www', ''], ['www,']],
         ["http://user:pass@www.do-mai_n.nl/path/name.ext",
             ['http', 'user', '<unk>', 'www', 'do-mai_n', 'nl', 'path', 'name', 'ext'],
             ['http', 'user', '<unk>', 'www', 'do-mai_n', 'nl', 'path', 'name', 'ext'],
             ['http://user:pass@www.do-mai_n.nl/path/name.ext']],
        ]

    # Low-level regex pattern tests
    # Important are the text and the number of resulting list elements,
    # less so the exact distribution of whitespace.
    sentence_pattern_tests =[
        ["s1", ["s1"]],
        ["s1.", ["s1."]],
        ["s1. ", ["s1.", " "]],
        ["s1. s2. ", ["s1.", " s2.", " "]],
        ["<bot:txt> s1. s2. ", ['<bot:txt> s1.', ' s2.', ' ']],

        ["s1\n", ["s1\n"]],
        ["s1. \n", ["s1.", " \n"]],

        ["s1\n\n", ["s1\n", "\n"]],
        ["s1\n \n", ["s1\n ", "\n"]],
        ["s1 \n \n", ["s1 \n ", "\n"]],

        ["s1\n\n\n\n\n", ["s1\n\n\n\n", "\n"]],

        ["s1.\ns2. ", ["s1.", "\ns2.", " "]],
        ["s1.\ns2.\ns3. ", ["s1.", "\ns2.", "\ns3.", " "]],

        ["s1. \ns2. ", ["s1.", " \ns2.", " "]],
        ["s1. \ns2. \ns3. ", ["s1.", " \ns2.", " \ns3.", " "]],

        ["s1. s2 <s> s3\n\n", ['s1.', ' s2 <s>', ' s3\n', '\n']],
        ["s1. s2 <s> s3\n\ns4", ['s1.', ' s2 <s>', ' s3\n', '\ns4']],
    ]

    suites = []

    suite = unittest.TestSuite()
    for i,a in enumerate(tests):
        suite.addTest(_TestTokenization('test_tokenize_text', a[0], a[1]))
        suite.addTest(_TestTokenization('test_tokenize_context', a[0], a[2]))
        suite.addTest(_TestTokenization('test_split_sentences', a[0], a[3]))
    suites.append(suite)

    suite = unittest.TestSuite()
    for i,a in enumerate(sentence_pattern_tests):
        suite.addTest(_TestPatterns('test_sentence_pattern', a[0], a[1]))
    suites.append(suite)

    suite = unittest.TestSuite()
    for _class in [_TestMultiOrderMisc,
                   _TestMultiOrderRemove,
                   _TestMultiOrderLoadingRobustness]:
        test_methods = unittest.TestLoader().getTestCaseNames(_class)
        for order in range(2, 5+1):
            for method in test_methods:
                suite.addTest(_class(method, order))
    suites.append(suite)

    suite = unittest.TestLoader().loadTestsFromTestCase(_TestModel)
    suites.append(suite)

    alltests = unittest.TestSuite(suites)
    return alltests


def test():
    runner = unittest.TextTestRunner(verbosity=1)
    runner.run(suite())

#class _TestSuiteAllTests(unittest.TestSuite):
#    def __init__(self):
#        self.add(suite())

if __name__ == '__main__':
    unittest.main()





#!/usr/bin/python3

# Copyright © 2012 marmuta <marmvta@gmail.com>
#
# This file is part of Onboard.
#
# Onboard is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Onboard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import subprocess
import re
import tempfile
from glob import glob
import gettext
import unittest


class TestTranslations(unittest.TestCase):

    def setUp(self):
        """ create mo files so we can read them with gettext later """
        self.tmp_dir = tempfile.TemporaryDirectory(prefix="test_onboard_")
        localedir = self.tmp_dir.name

        files = glob("po/*.po")
        if not files:
            Exception("No po files found, aborting")
        for file in files:
            name = os.path.splitext(os.path.basename(file))[0]
            dstfile = os.path.join(localedir, name, "LC_MESSAGES", "onboard.mo")
            os.makedirs(os.path.dirname(dstfile))
            subprocess.call(["msgfmt", file, "-o", dstfile])

    #@unittest.expectedFailure
    def test_field_names(self):
        """ compare field names of all msgids and msgstrs for all languages """
        localedir = self.tmp_dir.name
        files = glob(os.path.join(localedir, "*", "LC_MESSAGES", "onboard.mo"))
        langs = [file.split(os.path.sep)[-3] for file in files]
        totalstrcount = 0
        strcount = 0
        totallangcount = 0
        langcount = 0
        msg = ""

        for lang in langs:
            trans = gettext.translation("onboard", localedir, [lang])
            pattern = re.compile("{[^}]*}")
            count = 0
            for msgid, msgstr in trans._catalog.items():
                idfields = pattern.findall(msgid)
                strfields = pattern.findall(msgstr)
                if idfields != strfields:
                    msg += "\nField mismatch: Lang={} msgid='{}' msgstr='{}'" \
                           .format(lang, msgid, msgstr)
                    count += 1
                totalstrcount += 1
            totallangcount += 1

            if count:
                langcount += 1
            strcount += count

        self.assertEqual(strcount, 0, "{}/{} strings in {}/{} language(s) "
                                      "with non-matching field names\n"
                                      "{}" \
                                      .format(strcount, totalstrcount,
                                              langcount, totallangcount, msg))

